import copyreg
import pickle

import boost_histogram as bh
import h5py
import hdf5plugin
import hist
import numpy as np

MIN_PROTOCOL_VERSION = 1
CURRENT_PROTOCOL_VERSION = 1


class H5PickleProxy:
    """Allows storage of objects in h5py files with lazy reading."""

    def __init__(self, obj, h5group=None):
        self.obj = obj
        self.h5group = h5group

    def get(self):
        if self.obj is None:
            if self.h5group is None:
                raise ValueError(
                    "Trying to read object from H5PickleProxy but neither object nor underlying hdf5 storage is available."
                )
            elif not self.h5group:
                raise ValueError(
                    "Trying to read object from H5PickleProxy but the underlying file has been closed."
                )

            self.obj = pickle_load_h5py(self.h5group)

        return self.obj

    def release(self):
        self.obj = None


class H5Buffer:
    """Signals out-of-band storage of buffers in h5py files during pickling.

    The corresponding data can later be read back directly from the file using readinto.
    The provides an efficient mechanism for restoring objects which are not able to adopt
    external buffers.
    """

    def __init__(self, buf, h5dset=None, readonly=None):
        self.buf = buf
        self.h5dset = h5dset
        self.readonly = readonly

        if self.buf is not None:
            self.readonly = memoryview(self.buf).readonly

    def checkdset(self):
        if self.h5dset is None:
            raise ValueError(
                "Trying to read from H5Buffer but neither buffer nor dset are available."
            )
        elif not self.h5dset:
            raise ValueError(
                "Trying to read from H5Buffer but underlying file has been closed."
            )

    def getbuffer(self):
        if self.buf is None:
            self.checkdset()
            bufarr = np.empty_like(self.h5dset)
            if self.h5dset.size:
                self.h5dset.read_direct(bufarr)
            bufout = bufarr.data
            if self.readonly:
                bufout = bufout.toreadonly()
            self.buf = bufout

        return self.buf

    def readinto(self, b, /):
        # erase shape, strides and format information of the destination buffer
        bufarr = np.frombuffer(pickle.PickleBuffer(b).raw(), dtype=np.ubyte)
        if self.buf is None:
            # no source buffer available, read directly from the h5py dataset
            self.checkdset()
            if self.h5dset.size:
                self.h5dset.read_direct(bufarr)
            else:
                # read_direct doesn't work with empty datasets so fall back to slicing.
                # although this is a null operation it acts as a consistency check on
                # the destination buffer size'
                bufarr[...] = self.h5dset[...]
        else:
            # read from the stored buffer, erasing shape, strides and format information
            bufarr[...] = pickle.PickleBuffer(self.buf).raw()

    def __reduce_ex__(self, protocol):
        if protocol >= 5:
            outbuf = pickle.PickleBuffer(self.buf)
        else:
            if memoryview(self.buf).readonly:
                outbuf = bytes(self.buf)
            else:
                outbuf = bytearray(self.buf)

        return (type(self), (outbuf,))


class H5Path:
    """Lightweight softlink object storing the path to an h5py group or dataset within an hdf5 file."""

    def __init__(self, path):
        self.path = path


# functions below are a workaround for suboptimal pickling of bh.Histogram
# which doesn't properly use the out-of-band-mechanism and which can't adopt external buffers
def get_histogram_view(h):
    """Return a view to the contents of a histogram, suitable for use with PickleBuffer or H5Buffer."""
    return h.view(flow=True)


def make_Histogram(axes, storage, metadata, h5buf):
    """Reconstruct a Histogram given constructor arguments and an io-like object with a readinto method."""
    hist_read = bh.Histogram(*axes, storage=storage, metadata=metadata)
    hist_read_view = get_histogram_view(hist_read)
    h5buf.readinto(hist_read_view)
    return hist_read


def reduce_Histogram(obj):
    """Custom reduction function for Histogram storing the constructor arguments and an H5Buffer of the contents."""
    axes = tuple(obj.axes)
    view = get_histogram_view(obj)
    h5buf = H5Buffer(view)

    return (make_Histogram, (axes, obj.storage_type(), obj.metadata, h5buf))


def make_Hist(axes, storage, metadata, label, name, h5buf):
    """Reconstruct a Hist given constructor arguments and an io-like object with a readinto method."""
    hist_read = hist.Hist(
        *axes, storage=storage, metadata=metadata, label=label, name=name
    )
    hist_read_view = get_histogram_view(hist_read)
    h5buf.readinto(hist_read_view)
    return hist_read


def reduce_Hist(obj):
    """Custom reduction function for Hist storing the constructor arguments and an H5Buffer of the contents."""
    axes = tuple(obj.axes)
    view = get_histogram_view(obj)
    h5buf = H5Buffer(view)

    return (
        make_Hist,
        (axes, obj.storage_type(), obj.metadata, obj.label, obj.name, h5buf),
    )


class H5IO:
    """Provides and io-like interface for reading and writing to a resizable h5py dataset."""

    def __init__(self, name, h5out, mode="r"):
        if mode == "r":
            self.dset = h5out[name]
        elif mode == "w":
            self.dset = h5out.create_dataset(
                name,
                shape=(0,),
                dtype=np.ubyte,
                maxshape=(None,),
                chunks=(1024 * 1024,),
                **hdf5plugin.Blosc(cname="lz4"),
            )
        else:
            raise ValueError("Unsupported mode.")
        self.pos = 0

    def write(self, b, /):
        if type(b) is bytes:
            flat_view = memoryview(b)
        else:
            # erase shape, strides and format information
            flat_view = pickle.PickleBuffer(b).raw()

        oldsize = self.dset.size
        nbytes = flat_view.nbytes
        size = oldsize + nbytes
        self.dset.resize((size,))
        self.dset[oldsize:] = flat_view
        return nbytes

    def read(self, size=-1, /):
        oldpos = self.pos
        self.pos += size
        return self.dset[oldpos : self.pos].data

    def readline(self, size=-1, /):
        raise NotImplementedError("readline is not implemented")

    def readinto(self, b, /):
        # erase shape, strides and format information of the destination buffer
        arr = np.frombuffer(pickle.PickleBuffer(b).raw(), dtype=np.ubyte)
        nbytes = arr.nbytes
        oldpos = self.pos
        self.pos += nbytes
        if nbytes:
            self.dset.read_direct(arr, source_sel=np.s_[oldpos : self.pos])
        else:
            # read_direct doesn't work with zero bytes so fall back to slicing.
            # although this is a null operation it acts as a consistency check on
            # the destination buffer size'
            arr[...] = self.dset[...]

        return nbytes


class H5Pickler(pickle.Pickler):
    """Implements pickling to h5py file.

    Includes support for additional out-of-band storage for proxy objects and buffers which need
    to be read back directly from the file.  h5py groups and datasets can also be stored via H5Path softlink
    """

    # use custom dispatch table with custom reduce functions for boost_histogram.Histogram and hist.Hist
    dispatch_table = copyreg.dispatch_table.copy()
    dispatch_table[bh.Histogram] = reduce_Histogram
    dispatch_table[hist.hist.Hist] = reduce_Hist

    def __init__(self, h5group, protocol=5):
        self.h5group = h5group
        self.io = H5IO("pickle_data", h5group, mode="w")
        super().__init__(self.io, protocol=protocol)

    def reduce_H5PickleProxy(self, proxy):
        """Custom reduce function for H5PickleProxy.

        The proxied object is written out-of-band from the main pickle stream.
        The H5PickleProxy object is read back with only the group object where the proxied object can be lazily read back from.
        """

        PROXIED_OBJECTS_NAME = "proxied_objects"
        current_location = self.h5group

        if PROXIED_OBJECTS_NAME in current_location:
            proxied_objects = current_location[PROXIED_OBJECTS_NAME]
        else:
            proxied_objects = current_location.create_group(
                PROXIED_OBJECTS_NAME, track_order=True
            )

        iproxy = len(proxied_objects)
        proxied_group = pickle_dump_h5py(
            f"proxied_object_{iproxy}", proxy.obj, proxied_objects
        )

        return (type(proxy), (None, proxied_group))

    def reduce_H5Buffer(self, h5buf):
        """Custom reduce function for H5Buffer.

        The corresponding buffer is written out-of-band from the main pickle stream.
        The H5Buffer object is read back with only the dataset object where the underlying data can
        be read back when needed (and directly into an existing buffer if desired).
        """

        H5BUFFER_GROUP_NAME = "h5_buffers"
        current_location = self.h5group

        if H5BUFFER_GROUP_NAME in current_location:
            h5_buffers = current_location[H5BUFFER_GROUP_NAME]
        else:
            h5_buffers = current_location.create_group(
                H5BUFFER_GROUP_NAME, track_order=True
            )

        ibuf = len(h5_buffers)
        flat_view = pickle.PickleBuffer(h5buf.getbuffer()).raw()
        bufdset = h5_buffers.create_dataset(
            f"h5_buffer_{ibuf}",
            data=flat_view,
            chunks=True,
            **hdf5plugin.Blosc(cname="lz4"),
        )

        return (type(h5buf), (None, bufdset, flat_view.readonly))

    def reducer_override(self, obj):
        if type(obj) is H5PickleProxy:
            return self.reduce_H5PickleProxy(obj)
        elif type(obj) is H5Buffer:
            return self.reduce_H5Buffer(obj)
        else:
            return NotImplemented

    def persistent_id(self, obj):
        """Handle storage of h5py group or dataset as H5Path."""
        if isinstance(obj, h5py.Group) or isinstance(obj, h5py.Dataset):
            return H5Path(obj.name)
        else:
            return None


class H5Unpickler(pickle.Unpickler):
    """Implements unpickling from an h5py file.

    Includes support for reading back references to h5py groups and datasets stored as H5Path softlinks.
    """

    def __init__(self, h5group):
        self.h5group = h5group
        self.io = H5IO("pickle_data", h5group, mode="r")
        super().__init__(self.io)

    def persistent_load(self, pid):
        """Handle retrieval of h5py group or dataset via H5Path."""
        if type(pid) is H5Path:
            return self.h5group.file[pid.path]
        else:
            raise pickle.UnpicklingError("unsupported persistent object")


def pickle_dump_h5py(name, obj, h5out):
    """Write an object to a new h5py group which will be created within the provided group."""
    obj_group = h5out.create_group(name)
    try:
        obj_group.attrs["narf_h5py_pickle_protocol_version"] = CURRENT_PROTOCOL_VERSION
        pickler = H5Pickler(obj_group)
        pickler.dump(obj)
        return obj_group
    except:
        del obj_group
        raise


def pickle_load_h5py(h5group):
    """Load an object from the h5py group which contains it."""
    try:
        narf_protocol_version = h5group.attrs["narf_h5py_pickle_protocol_version"]
    except KeyError:
        raise ValueError(
            "h5py dataset does not contain a python object pickled by narf."
        )

    if narf_protocol_version < MIN_PROTOCOL_VERSION:
        raise ValueError(
            f"Unsuppported narf protocol version {narf_protocol_version}, minimum supported version is {MIN_PROTOCOL_VERSION}"
        )
    elif narf_protocol_version > CURRENT_PROTOCOL_VERSION:
        raise ValueError(
            f"Unsuppported narf protocol version {narf_protocol_version}, maximum supported version is {CURRENT_PROTOCOL_VERSION}"
        )

    unpickler = H5Unpickler(h5group)
    return unpickler.load()



import logging
import os
import time


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""

    green = "\x1b[1;32m"
    grey = "\x1b[38;21m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    # format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    myformat = "%(levelname)s:%(filename)s: %(message)s"

    FORMATS = {
        logging.DEBUG: green + myformat + reset,
        logging.INFO: grey + myformat + reset,
        logging.WARNING: yellow + myformat + reset,
        logging.ERROR: red + myformat + reset,
        logging.CRITICAL: bold_red + myformat + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


logging_verboseLevel = [
    logging.CRITICAL,
    logging.ERROR,
    logging.WARNING,
    logging.INFO,
    logging.DEBUG,
]


def set_logging_level(log, verbosity):
    log.setLevel(logging_verboseLevel[max(0, min(4, verbosity))])


def setup_logger(basefile, verbosity=3, no_colors=False, initName="wremnants"):

    setup_func = setup_base_logger if no_colors else setup_color_logger
    logger = setup_func(os.path.basename(basefile), verbosity, initName)
    # count messages of base logger
    base_logger = logging.getLogger("wremnants")
    add_logging_counter(base_logger)
    # stop total time
    add_time_info("Total time")
    return logger


def setup_color_logger(name, verbosity, initName="wremnants"):
    base_logger = logging.getLogger(initName)
    # set console handler
    ch = logging.StreamHandler()
    ch.setFormatter(CustomFormatter())
    base_logger.addHandler(ch)
    set_logging_level(base_logger, verbosity)
    base_logger.propagate = False  # to avoid propagating back to root logger, which would print messages twice
    return base_logger.getChild(name)


def setup_base_logger(name, verbosity, initName="wremnants"):
    logging.basicConfig(format="%(levelname)s: %(message)s")
    base_logger = logging.getLogger(initName)
    set_logging_level(base_logger, verbosity)
    return base_logger.getChild(name)


def child_logger(name, initName="wremnants"):
    # count messages of child logger
    logger = logging.getLogger(initName).getChild(name)
    add_logging_counter(logger)
    return logger


class LoggingCounterHandler(logging.Handler):
    def __init__(self, level):
        super().__init__()
        self.count = 0
        self.level = level

    def emit(self, record):
        if record.levelno == self.level:
            self.count += 1


def add_logging_counter(
    logger, levels=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
):
    logger.counter = {}
    for level in levels:
        logger.counter[level] = LoggingCounterHandler(getattr(logging, level))
        logger.addHandler(logger.counter[level])


def print_logging_count(logger, verbosity=logging.WARNING):
    if hasattr(logger, "counter"):
        for level, counter in logger.counter.items():
            if getattr(logging, level) < verbosity:
                continue
            if counter.count > 0:
                getattr(logger, level.lower())(
                    f"Logger {logger.name} triggered {level} messages: {counter.count}"
                )


def add_time_info(tag, logger=logging.getLogger("wremnants")):
    if not hasattr(logger, "times"):
        logger.times = {}
    logger.times[tag] = time.time()


def print_time_info(logger):
    if hasattr(logger, "times"):
        time_end = time.time()
        logger.info(f"Time tags summary {logger.name}:")
        for tag, itime in logger.times.items():
            logger.info(f"{tag}: {time_end - itime}")


def summary(verbosity=logging.WARNING, extended=True):
    base_logger = logging.getLogger("wremnants")

    base_logger.info(f"--------------------------------------")
    base_logger.info(f"----------- logger summary -----------")
    base_logger.info(f"--------------------------------------")

    print_time_info(base_logger)

    print_logging_count(base_logger, verbosity=verbosity)

    if not extended:
        return

    # Iterate through all child loggers and print their names, levels, and counts
    all_loggers = logging.Logger.manager.loggerDict
    for logger_name, logger_obj in all_loggers.items():
        if logger_name.startswith("wremnants."):
            print_logging_count(logger_obj, verbosity=verbosity)



import copy
import itertools
from functools import reduce

import hist
import numpy as np

from wums import logging

logger = logging.child_logger(__name__)


def valsAndVariances(h1, h2, flow=True):
    return (
        h1.values(flow=flow),
        h2.values(flow=flow),
        h1.variances(flow=flow),
        h2.variances(flow=flow),
    )


# Broadcast h1 to match the shape of h2
def broadcastSystHist(h1, h2, flow=True, by_ax_name=True):
    if h1.ndim > h2.ndim or h1.shape == h2.shape:
        return h1

    s1 = h1.values(flow=flow).shape
    s2 = h2.values(flow=flow).shape

    # the additional axes have to be broadcasted as leading
    # Either do this by name, or by broadcasting from the right (numpy default broadcasts from left)
    if by_ax_name:
        moves = {
            i: e
            for i, (e, n2) in enumerate(zip(s2, h2.axes.name))
            if n2 not in h1.axes.name
        }
    else:
        moves = {
            h2.ndim - 1 - i: h2.values(flow=flow).shape[h2.ndim - 1 - i]
            for i in range(h2.ndim - h1.ndim)
        }

    broadcast_shape = list(moves.values()) + list(s1)

    try:
        new_vals = np.broadcast_to(h1.values(flow=flow), broadcast_shape)
    except ValueError:
        raise ValueError(
            "Cannot broadcast hists with incompatible axes!\n"
            f"    h1.shape {h1.shape}; h2.shape: {h2.shape}\n"
            f"    h1.axes: {h1.axes}\n"
            f"    h2.axes: {h2.axes}"
        )

    # move back to original order
    new_vals = np.moveaxis(new_vals, np.arange(len(moves)), list(moves.keys()))

    if new_vals.shape != h2.values(flow=flow).shape:
        raise ValueError(
            f"Broadcast shape {new_vals.shape} (from h1.shape={h1.values(flow=flow).shape}, axes={h1.axes.name}) "
            f"does not match desired shape {h2.view(flow=flow).shape} (axes={h2.axes.name})"
        )

    if h1.storage_type == hist.storage.Weight:
        new_vars = np.broadcast_to(h1.variances(flow=flow), broadcast_shape)
        new_vars = np.moveaxis(new_vars, np.arange(len(moves)), list(moves.keys()))
        new_vals = np.stack((new_vals, new_vars), axis=-1)
    return hist.Hist(*h2.axes, data=new_vals, storage=h1.storage_type())


def divideHists(
    h1,
    h2,
    cutoff=1e-5,
    allowBroadcast=True,
    rel_unc=False,
    cutoff_val=1.0,
    flow=True,
    createNew=True,
    by_ax_name=True,
):
    """
    Parameters:
    rel_unc (bool): Treat the divisor as a constant

    Returns:
    hist: h1/h2
    """
    if allowBroadcast:
        h1 = broadcastSystHist(h1, h2, flow, by_ax_name)
        h2 = broadcastSystHist(h2, h1, flow, by_ax_name)

    h1vals, h2vals, h1vars, h2vars = valsAndVariances(h1, h2, flow=flow)

    if rel_unc and h1vars is not None:
        storage = hist.storage.Weight()
    elif h1vars is None or h2vars is None:
        storage = hist.storage.Double()
    else:
        storage = hist.storage.Weight()

    outh = hist.Hist(*h1.axes, storage=storage) if createNew else h1

    # Careful not to overwrite the values of h1
    out = (
        outh.values(flow=flow)
        if createNew
        else np.full(outh.values(flow=flow).shape, cutoff_val)
    )

    # Apply cutoff to both numerator and denominator
    # only do the division if denominator is above cutoff, keep nominator values else
    cutoff_criteria = np.abs(h2vals) > cutoff

    # if denominator and nominator below cutoff, set to cutoff_val. By the argument that 0/0 = 1
    cutoff_criteria2 = (np.abs(h2vals) < cutoff) & (np.abs(h1vals) < cutoff)
    out[cutoff_criteria2] = cutoff_val
    val = np.divide(h1vals, h2vals, out=out, where=cutoff_criteria)

    if cutoff_criteria2.sum():
        logger.warning(
            f"Encountered {cutoff_criteria2.sum()} values below {cutoff} in h1 and h2 in divideHists, these will be set to {cutoff_val}"
        )
    elif (~cutoff_criteria).sum():
        logger.warning(
            f"Encountered {(~cutoff_criteria).sum()} values below {cutoff} in h2 in divideHists, these will be set to h1"
        )

    if outh.storage_type == hist.storage.Weight:
        rel1 = relVariance(h1vals, h1vars, cutoff=cutoff)
        val2 = np.multiply(val, val)
        if rel_unc:
            var = np.multiply(val2, rel1, out=val2)
        else:
            rel2 = relVariance(h2vals, h2vars, cutoff=cutoff)
            relsum = np.add(rel1, rel2)
            var = np.multiply(relsum, val2, out=val2)

        outh.view(flow=flow)[...] = np.stack((val, var), axis=-1)
    else:
        outh.values(flow=flow)[...] = val

    return outh


def relVariance(hvals, hvars, cutoff=1e-5, fillOnes=False):
    out = np.empty(hvars.shape)
    np.multiply(hvals, hvals, out=out)
    np.clip(out, a_min=cutoff, a_max=None, out=out)
    np.divide(hvars, out, out=out)
    if fillOnes:
        out[hvals < cutoff] = 1.0
    return out


def relVariances(h1vals, h2vals, h1vars, h2vars, cutoff=1e-5):
    rel1 = relVariance(h1vals, h1vars, cutoff=cutoff)
    rel2 = relVariance(h2vals, h2vars, cutoff=cutoff)
    return (rel1, rel2)


# TODO: Implement this rather than relying on pdf unc function
# def rssHist(h):


def sqrtHist(h):
    rootval = np.sqrt(h.values(flow=True))
    rooth = h.copy()

    if h.storage_type == hist.storage.Double:
        rooth[...] = rootval
    else:
        relvar = relVariance(h.values(flow=True), h.variances(flow=True))
        newvar = 0.5 * rootval * rootval * relvar
        rooth[...] = np.stack((rootval, newvar), axis=-1)

    return rooth


def multiplyWithVariance(vals1, vals2, vars1=None, vars2=None):
    outvals = np.empty(vals1.shape)
    outvars = (
        np.empty(vars1.shape) if (vars1 is not None and vars2 is not None) else None
    )
    np.multiply(vals1, vals2, out=outvals)
    if outvars is not None:
        np.multiply(outvals, outvals, out=outvars)
        relvar1, relvar2 = relVariances(vals1, vals2, vars1, vars2)
        np.add(relvar1, relvar2, out=relvar1)
        np.multiply(outvars, relvar1, out=outvars)

    return outvals, outvars


def multiplyHists(h1, h2, allowBroadcast=True, createNew=True, flow=True):
    if allowBroadcast:
        h1 = broadcastSystHist(h1, h2, flow=flow)
        h2 = broadcastSystHist(h2, h1, flow=flow)

    if (
        h1.storage_type == hist.storage.Double
        and h2.storage_type == hist.storage.Double
        and createNew == False
    ):
        return h1 * h2

    with_variance = (
        h1.storage_type == hist.storage.Weight
        and h2.storage_type == hist.storage.Weight
    )
    outh = h1
    vals, varis = multiplyWithVariance(
        h1.values(flow=flow),
        h2.values(flow=flow),
        h1.variances(flow=flow) if with_variance else None,
        h2.variances(flow=flow) if with_variance else None,
    )

    if createNew:
        outh = hist.Hist(*outh.axes, storage=outh.storage_type())
    outh.values(flow=flow)[...] = vals
    if varis is not None:
        outh.variances(flow=flow)[...] = varis

    return outh


# TODO: Figure out the overflow
def concatenateHists(h1, h2, allowBroadcast=True, by_ax_name=True, flow=False):
    if allowBroadcast:
        h1 = broadcastSystHist(h1, h2, flow=flow, by_ax_name=by_ax_name)
        h2 = broadcastSystHist(h2, h1, flow=flow, by_ax_name=by_ax_name)

    axes = []
    for ax1, ax2 in zip(h1.axes, h2.axes):
        if ax1 == ax2:
            axes.append(ax1)
            continue

        if type(ax1) != type(ax2):
            raise ValueError("Cannot combine inconsistent axis types!")

        if type(ax1) == hist.axis.StrCategory:
            new_entries = list(ax1) + list(ax2)
            if len(new_entries) != len(set(new_entries)):
                raise ValueError(
                    "Cannot concatenate StrCategory axes with duplicate labels"
                )
            axes.append(hist.axis.StrCategory(new_entries, name=ax1.name))
        else:
            if ax1.edges[0] > ax2.edges[0]:
                ax1, ax2 = ax2, ax1

            if ax1.edges[-1] == ax2.edges[0]:
                axes.append(
                    hist.axis.Variable(
                        np.concatenate((ax1.edges, ax2.edges[1:])),
                        name=ax1.name,
                        underflow=ax1.traits.underflow,
                        overflow=ax2.traits.overflow,
                    )
                )
            else:
                raise ValueError(
                    f"Cannot concatenate hists with inconsistent axes: {ax1.name} and {ax2.name}"
                )

    newh = hist.Hist(*axes, storage=h1.storage_type())
    fill_by_centers(newh, h1)
    fill_by_centers(newh, h2)

    return newh


def concatenate_syst_hists(hists, has_nominal=True):
    if len(hists) < 2:
        raise ValueError("At least two hists required to concatenate")
    hcat = hists[0]
    for h in hists[1:]:
        hcat = concatenateHists(hcat, h[..., has_nominal:], allowBroadcast=False)

    return hcat


def fill_by_centers(h, href):
    to_fill = [
        list(y) if type(y) == hist.axis.StrCategory else x
        for x, y in zip(href.axes.centers, href.axes)
    ]
    h.view()[h.axes.index(*to_fill)] = href.view()
    return h


def addHists(
    h1,
    h2,
    allowBroadcast=True,
    createNew=True,
    scale1=None,
    scale2=None,
    flow=True,
    by_ax_name=True,
):
    if allowBroadcast:
        h1 = broadcastSystHist(h1, h2, flow=flow, by_ax_name=by_ax_name)
        h2 = broadcastSystHist(h2, h1, flow=flow, by_ax_name=by_ax_name)

    h1vals, h2vals, h1vars, h2vars = valsAndVariances(h1, h2, flow=flow)
    hasWeights = (
        h1._storage_type() == hist.storage.Weight()
        and h2._storage_type() == hist.storage.Weight()
    )
    # avoid scaling the variance if not needed, to save some time
    # I couldn't use hvals *= scale, otherwise I get this error: ValueError: output array is read-only
    if scale1 is not None:
        h1vals = scale1 * h1vals
        if hasWeights:
            h1vars = (scale1 * scale1) * h1vars
    if scale2 is not None:
        h2vals = scale2 * h2vals
        if hasWeights:
            h2vars = (scale2 * scale2) * h2vars

    outh = h1
    if createNew:
        if not hasWeights:
            return hist.Hist(*outh.axes, data=h1vals + h2vals)
        else:
            return hist.Hist(
                *outh.axes,
                storage=hist.storage.Weight(),
                data=np.stack((h1vals + h2vals, h1vars + h2vars), axis=-1),
            )
    else:
        outvals = h1vals if h1.shape == outh.shape else h2vals
        np.add(h1vals, h2vals, out=outvals)
        outh.values(flow=True)[...] = outvals
        if hasWeights:
            outvars = h1vars if h1.shape == outh.shape else h2vars
            np.add(h1vars, h2vars, out=outvars)
            outh.variances(flow=True)[...] = outvars
        elif h1._storage_type() == hist.storage.Weight():
            logger.debug(
                "Histogram h1 has weights but h2 not, values for h1 are updated but variances not."
            )

        return outh


def sumHists(hists):
    return reduce(addHists, hists)


def mirrorHist(hvar, hnom, cutoff=1):
    div = divideHists(hnom, hvar, cutoff, createNew=True)
    hnew = multiplyHists(div, hnom, createNew=False)
    return hnew


def extendHistByMirror(hvar, hnom, downAsUp=False, downAsNomi=False):
    hmirror = mirrorHist(hvar, hnom)
    if downAsNomi:
        hvar, hmirror = hmirror, hvar

    mirrorAx = hist.axis.Integer(0, 2, name="mirror", overflow=False, underflow=False)

    if (
        hvar.storage_type == hist.storage.Weight
        and hnom.storage_type == hist.storage.Weight
    ):
        hnew = hist.Hist(
            *hvar.axes,
            mirrorAx,
            storage=hist.storage.Weight(),
            data=np.stack((hvar.view(flow=True), hmirror.view(flow=True)), axis=-1),
        )
    else:
        hnew = hist.Hist(
            *hvar.axes,
            mirrorAx,
            data=np.stack((hvar.values(flow=True), hmirror.values(flow=True)), axis=-1),
        )

    return hnew


# add new axis and set values of old histogram to idx
def addGenChargeAxis(h, idx):
    return addGenericAxis(
        h,
        hist.axis.Regular(2, -2.0, 2.0, underflow=False, overflow=False, name="qGen"),
        idx,
        add_trailing=False,
        flow=True,
    )


def addSystAxis(h, size=1, offset=0, axname="systIdx"):
    return addGenericAxis(
        h, hist.axis.Regular(size, offset, size + offset, name=axname)
    )


def addGenericAxis(h, axis, idx=None, add_trailing=True, flow=True):
    axes = [*h.axes, axis] if add_trailing else [axis, *h.axes]
    hnew = hist.Hist(*axes, storage=h.storage_type())
    if idx != None:
        # add old histogram only in single bin
        slices = [idx if ax == axis else slice(None) for ax in hnew.axes]
        hnew.view(flow=flow)[tuple(slices)] = h.view(flow=flow)
    else:
        # Broadcast to new shape
        slices = [np.newaxis if ax == axis else slice(None) for ax in hnew.axes]
        hnew.view(flow=flow)[...] = hnew.view(flow=flow) + h.view(flow=flow)[tuple(slices)]
    return hnew


def clipNegativeVals(h, clipValue=0, createNew=False):
    newh = h.copy() if createNew else h
    np.clip(
        newh.values(flow=True), a_min=clipValue, a_max=None, out=newh.values(flow=True)
    )
    return newh


def scaleHist(h, scale, createNew=True, flow=True, scaleVarianceLinearly=False):
    varianceScale = scale if scaleVarianceLinearly else scale * scale
    if createNew:
        if h.storage_type == hist.storage.Double:
            hnew = hist.Hist(*h.axes)
        else:
            hnew = hist.Hist(*h.axes, storage=hist.storage.Weight())
            hnew.variances(flow=flow)[...] = varianceScale * h.variances(flow=flow)

        hnew.values(flow=flow)[...] = scale * h.values(flow=flow)

        return hnew
    else:
        h.values(flow=flow)[...] *= scale
        if h.storage_type == hist.storage.Weight:
            h.variances(flow=flow)[...] *= varianceScale
        return h


def normalize(h, scale=1e6, createNew=True, flow=True):
    if h.storage_type == hist.storage.Weight:
        scale = scale / h.sum(flow=flow).value
    else:
        scale = scale / h.sum(flow=flow)
    return scaleHist(h, scale, createNew, flow)


def makeAbsHist(h, axis_name, rename=True):
    ax = h.axes[axis_name]
    axidx = list(h.axes).index(ax)
    axInfo = dict(
        underflow=False,
        overflow=ax.traits.overflow,
        name=f"abs{axis_name}" if rename else axis_name,
    )
    if ax.size == 1 and -ax.edges[0] == ax.edges[-1]:
        abs_ax = hist.axis.Regular(1, 0, ax.edges[-1], **axInfo)
        return hist.Hist(
            *h.axes[:axidx],
            abs_ax,
            *h.axes[axidx + 1 :],
            storage=h.storage_type(),
            data=h.view(flow=True),
        )

    if 0 not in ax.edges:
        raise ValueError("Can't mirror around 0 if it isn't a bin boundary")
    new_edges = ax.edges[ax.index(0.0) :]
    if isinstance(ax, hist.axis.Regular):
        abs_ax = hist.axis.Regular(len(new_edges) - 1, 0, new_edges[-1], **axInfo)
    else:
        abs_ax = hist.axis.Variable(new_edges, **axInfo)

    hnew = hist.Hist(
        *h.axes[:axidx], abs_ax, *h.axes[axidx + 1 :], storage=h.storage_type()
    )

    s = hist.tag.Slicer()
    view = h[{axis_name: s[ax.index(0) :]}].view(flow=True) + np.flip(
        h[{axis_name: s[: ax.index(0)]}].view(flow=True), axis=axidx
    )
    hnew[...] = np.take(
        view, range(ax.traits.underflow, abs_ax.size + ax.traits.underflow), axidx
    )
    return hnew


# Checks if edges1 could be rebinned to edges2. Order is important!
def compatibleBins(edges1, edges2):
    comparef = np.vectorize(lambda x: np.isclose(x, edges1).any())
    return np.all(comparef(edges2))


def rebinHistMultiAx(h, axes, edges=[], lows=[], highs=[]):
    # edges: lists of new edges or integers to merge bins, in case new edges are given the lows and highs will be ignored
    # lows: list of new lower boundaries
    # highs: list of new upper boundaries

    sel = {}
    for ax, low, high, rebin in itertools.zip_longest(axes, lows, highs, edges):
        if ax not in h.axes.name:
            logger.debug(f"Did not find axis {ax} in hist. Skipping rebin.")
            continue
        if rebin not in [None, []] and type(rebin) != int:
            h = rebinHist(h, ax, rebin)
        elif low is not None and high is not None:
            # in case high edge is upper edge of last bin we need to manually set the upper limit
            upper = hist.overflow if high == h.axes[ax].edges[-1] else complex(0, high)
            logger.info(f"Restricting the axis '{ax}' to range [{low}, {high}]")
            sel[ax] = slice(
                complex(0, low), upper, hist.rebin(rebin) if rebin else None
            )
        elif type(rebin) == int and rebin > 1:
            logger.info(f"Rebinning the axis '{ax}' by [{rebin}]")
            sel[ax] = slice(None, None, hist.rebin(rebin))
    return h[sel] if len(sel) > 0 else h


def mirrorAxis(h, axis, flow=True):
    # mirror an axis by projecting the entries to the absolute value of the axis and copy it
    htmp = makeAbsHist(h, axis, rename=False)
    idx = htmp.axes.name.index(axis)
    vals = htmp.values(flow=flow)
    varis = htmp.variances(flow=flow)
    vals_flip = np.flip(vals, axis=idx)
    varis_flip = np.flip(varis, axis=idx)
    vals = np.concatenate((vals_flip, vals), axis=idx)
    varis = np.concatenate((varis_flip, varis), axis=idx)
    h.values(flow=flow)[...] = vals / 2.0
    h.variances(flow=flow)[...] = varis / 4.0
    return h


def mirrorAxes(h, axes, flow=True):
    for a in axes:
        h = mirrorAxis(h, a, flow=flow)
    return h


def disableAxisFlow(ax):
    if isinstance(ax, hist.axis.Integer):
        args = [ax.edges[0], ax.edges[-1]]
    elif isinstance(ax, hist.axis.Regular):
        args = [ax.size, ax.edges[0], ax.edges[-1]]
    else:
        args = [ax.edges]

    return type(ax)(
        *args,
        name=ax.name,
        overflow=False,
        underflow=False,
        circular=ax.traits.circular,
    )


def disableFlow(h, axis_name):
    # disable the overflow and underflow bins of a single axes, while keeping the flow bins of other axes
    ax = h.axes[axis_name]
    ax_idx = [a.name for a in h.axes].index(axis_name)
    new_ax = disableAxisFlow(ax)
    axes = list(h.axes)
    axes[ax_idx] = new_ax
    hnew = hist.Hist(*axes, name=h.name, storage=h.storage_type())
    slices = [
        (
            slice(None)
            if i != ax_idx
            else slice(ax.traits.underflow, new_ax.size + ax.traits.underflow)
        )
        for i in range(len(axes))
    ]
    hnew.values(flow=True)[...] = h.values(flow=True)[tuple(slices)]
    if hnew.storage_type == hist.storage.Weight:
        hnew.variances(flow=True)[...] = h.variances(flow=True)[tuple(slices)]
    return hnew


def rebinHist(h, axis_name, edges, flow=True):
    if type(edges) == int:
        return h[{axis_name: hist.rebin(edges)}]

    ax = h.axes[axis_name]
    ax_idx = [a.name for a in h.axes].index(axis_name)

    if len(edges) == len(ax.edges) and np.isclose(edges, ax.edges).all():
        return h

    if not compatibleBins(ax.edges, edges):
        raise ValueError(
            f"Cannot rebin histogram due to incompatible edges for axis '{ax.name}'\n"
            f"Edges of histogram are {ax.edges}, requested rebinning to {edges}"
        )

    # If you rebin to a subset of initial range, keep the overflow and underflow
    overflow = ax.traits.overflow or (
        edges[-1] < ax.edges[-1] and not np.isclose(edges[-1], ax.edges[-1])
    )
    underflow = ax.traits.underflow or (
        edges[0] > ax.edges[0] and not np.isclose(edges[0], ax.edges[0])
    )
    new_ax = hist.axis.Variable(
        edges, name=ax.name, overflow=overflow, underflow=underflow
    )
    axes = list(h.axes)
    axes[ax_idx] = new_ax

    hnew = hist.Hist(*axes, name=h.name, storage=h.storage_type())

    if type(ax) == hist.axis.StrCategory:
        edge_idx = np.array(edges, dtype=int)
    else:
        # Offset from bin edge to avoid numeric issues
        offset = 0.5 * np.min(ax.edges[1:] - ax.edges[:-1])
        edges_eval = edges + offset
        edge_idx = ax.index(edges_eval)

    # Avoid going outside the range, reduceat will add the last index anyway
    if edge_idx[-1] == ax.size + ax.traits.overflow:
        edge_idx = edge_idx[:-1]

    if underflow:
        # Only if the original axis had an underflow should you offset
        if ax.traits.underflow:
            edge_idx += 1
        edge_idx = np.insert(edge_idx, 0, 0)

    # Take is used because reduceat sums i:len(array) for the last entry, in the case
    # where the final bin isn't the same between the initial and rebinned histogram, you
    # want to drop this value. Add tolerance of 1/2 min bin width to avoid numeric issues
    flow_add = (underflow + overflow) * flow
    hnew.values(flow=flow)[...] = np.add.reduceat(
        h.values(flow=flow), edge_idx, axis=ax_idx
    ).take(indices=range(new_ax.size + flow_add), axis=ax_idx)
    if hnew.storage_type == hist.storage.Weight:
        hnew.variances(flow=flow)[...] = np.add.reduceat(
            h.variances(flow=flow), edge_idx, axis=ax_idx
        ).take(indices=range(new_ax.size + flow_add), axis=ax_idx)
    return hnew


def get_rebin_actions(axes, ax_lim=[], ax_rebin=[], ax_absval=[], rename=False):
    if len(ax_lim) % 2 or len(ax_lim) / 2 > len(axes) or len(ax_rebin) > len(axes):
        raise ValueError(
            "Inconsistent rebin or axlim arguments. axlim must be at most two entries per axis, and rebin at most one"
        )
    actions = []
    if len(ax_lim) > 0 or len(ax_rebin) > 0:
        actions.append(
            lambda h, axes=axes, rebins=ax_rebin, lows=ax_lim[::2], highs=ax_lim[
                1::2
            ]: rebinHistMultiAx(h, axes, rebins, lows, highs)
        )

    for i, (var, absval) in enumerate(itertools.zip_longest(axes, ax_absval)):
        if absval:
            logger.info(f"Taking the absolute value of axis '{var}'")
            actions.append(lambda h, ax=var: makeAbsHist(h, ax, rename=rename))
            axes[i] = f"abs{var}" if rename else var
    return actions


def mergeAxes(ax1, ax2):
    if ax1.edges[0] < ax2.edges[0]:
        tmp = ax1
        ax1 = ax2
        ax2 = tmp

    if np.array_equal(ax1.edges, ax2.edges):
        return ax1

    ax1_edges = ax1.edges
    ax1_merge_idx = len(ax1.edges)
    while ax1_edges[ax1_merge_idx - 1] not in ax2.edges:
        ax1_merge_idx -= 1

    ax1_edges = ax1_edges[:ax1_merge_idx]
    if not ax1_edges.size:
        raise ValueError("Didn't find any common edges in two axes, can't merge")

    merge_idx = list(ax2.edges).index(ax1_edges[-1]) + 1
    if merge_idx < 1 or merge_idx > ax2.size + 1:
        raise ValueError(
            "Can't merge axes unless there is a common point of intersection! "
            f"Merge index was {merge_idx} "
            f"The edges were {ax1.edges} (size={ax1.size}), and {ax2.edges} (size={ax2.size})"
        )

    new_edges = np.concatenate((ax1_edges, ax2.edges[merge_idx:]))
    return hist.axis.Variable(new_edges, name=ax1.name)


def findAxes(hists, axis_idx):
    if type(axis_idx) in [list, tuple]:
        return [h.axes[[ax for ax in axis_idx if ax in h.axes.name][0]] for h in hists]
    else:
        return [h.axes[axis_idx] for h in hists]


def findCommonBinning(hists, axis_idx):
    if len(hists) < 2:
        raise ValueError("Can only find common binning between > 1 hists")

    orig_axes = findAxes(hists, axis_idx)
    # Set intersection with tolerance for floats
    common_edges = np.array(orig_axes[0].edges)
    for ax in orig_axes[1:]:
        common_edges = common_edges[
            np.isclose(np.array(ax.edges)[:, np.newaxis], common_edges).any(0)
        ]

    edges = np.sort(common_edges)
    logger.debug(f"Common edges are {common_edges}")

    if len(edges) < 2:
        raise ValueError(f"Found < 2 common edges, cannot rebin. Axes were {orig_axes}")
    return edges


def rebinHistsToCommon(hists, axis_idx, keep_full_range=False):
    orig_axes = findAxes(hists, axis_idx)
    new_edges = findCommonBinning(hists, axis_idx)
    rebinned_hists = [
        rebinHist(h, ax.name, new_edges) for h, ax in zip(hists, orig_axes)
    ]

    # TODO: This assumes that the range extension only happens in one direction,
    # specifically, that the longer range hist has higher values
    if keep_full_range:
        full_hists = []
        hists_by_max = sorted(hists, key=lambda x: x.axes[axis_idx].edges[-1])
        new_ax = rebinned_hists[0].axes[axis_idx]
        for h in hists_by_max:
            new_ax = mergeAxes(new_ax, h.axes[axis_idx])

        for h, rebinh in zip(hists, rebinned_hists):
            axes = list(rebinh.axes)
            axes[axis_idx] = new_ax
            newh = hist.Hist(*axes, name=rebinh.name, storage=h.storage_type())
            merge_idx = new_ax.index(rebinh.axes[axis_idx].edges[-1])
            # TODO: treat the overflow/underflow properly
            low_vals = rebinh.view()
            max_idx = min(
                h.axes[axis_idx].size,
                h.axes[axis_idx].index(newh.axes[axis_idx].edges[-1]),
            )
            vals = np.append(
                low_vals, np.take(h.view(), range(merge_idx, max_idx), axis_idx)
            )
            zero_pad_shape = list(newh.shape)
            zero_pad_shape[axis_idx] -= vals.shape[axis_idx]
            vals = np.append(
                vals,
                np.full(
                    zero_pad_shape,
                    (
                        hist.accumulators.WeightedSum(0, 0)
                        if newh.storage_type == hist.storage.Weight
                        else 0.0
                    ),
                ),
            )
            newh[...] = vals
            full_hists.append(newh)

        new_edges = findCommonBinning(full_hists, axis_idx)
        rebinned_hists = [
            rebinHist(h, h.axes[axis_idx].name, new_edges) for h in full_hists
        ]

    return rebinned_hists


def projectNoFlow(h, proj_ax, exclude=[]):
    s = hist.tag.Slicer()
    if type(proj_ax) == str:
        proj_ax = [proj_ax]
    hnoflow = h[
        {
            ax: s[0 : hist.overflow : hist.sum]
            for ax in h.axes.name
            if ax not in exclude + list(proj_ax)
        }
    ]
    return hnoflow.project(*proj_ax)


def extendEdgesByFlow(href, bin_flow_width=0.02):
    # add extra bin with bin wdith of a fraction of the total width
    all_edges = []
    for axis in href.axes:
        edges = axis.edges
        axis_range = edges[-1] - edges[0]
        if axis.traits.underflow:
            edges = np.insert(edges, 0, edges[0] - axis_range * bin_flow_width)
        if axis.traits.overflow:
            edges = np.append(edges, edges[-1] + axis_range * bin_flow_width)
        all_edges.append(edges)
    if len(all_edges) == 1:
        return all_edges[0]
    else:
        return all_edges


def unrolledHist(h, obs=None, binwnorm=None, add_flow_bins=False):
    # add_flow_bins to add the overflow and underflow bins into bins of the unrolled histogram
    if obs is not None:
        hproj = h.project(*obs)
    else:
        hproj = h

    if binwnorm:
        edges = (
            plot_tools.extendEdgesByFlow(hproj) if add_flow_bins else hproj.axes.edges
        )
        binwidths = np.outer(*[np.diff(e.squeeze()) for e in edges]).flatten()
        scale = binwnorm / binwidths
    else:
        scale = 1

    bins = np.prod(hproj.axes.extent) if add_flow_bins else np.prod(hproj.axes.size)
    newh = hist.Hist(
        hist.axis.Integer(0, bins, underflow=False, overflow=False),
        storage=hproj._storage_type(),
    )
    if hproj._storage_type() == hist.storage.Double():
        newh.view(flow=False)[...] = np.ravel(hproj.values(flow=add_flow_bins)) * scale
    else:
        newh.view(flow=False)[...] = np.stack(
            [
                np.ravel(hproj.values(flow=add_flow_bins)) * scale,
                np.ravel(hproj.variances(flow=add_flow_bins)) * scale**2,
            ],
            axis=-1,
        )
    return newh


def transfer_variances(h1, h2, flow=True):
    # take variances from h2 to set variances in h1 by scaling with the relative yields
    val2 = h2.values(flow=flow)
    var2 = h2.variances(flow=flow)
    val1 = h1.values(flow=flow)
    # extra dimensions for systematic axes
    extra_dimensions = (Ellipsis,) + (np.newaxis,) * (val1.ndim - val2.ndim)
    var1 = (
        np.ones_like(val1) * var2[extra_dimensions]
    )  # use var2 directly in cases of not finite values
    mask = np.isfinite(var2) & (var2 != 0.0) & (val2 != 0.0)
    # scale uncertainty with difference in yield with respect to second histogram such that relative uncertainty stays the same,
    factors = val1[mask] / val2[mask][extra_dimensions]
    var1[mask] = var2[mask][extra_dimensions] * factors**2
    hNew = hist.Hist(*h1.axes, storage=hist.storage.Weight())
    hNew.view(flow=flow)[...] = np.stack((val1, var1), axis=-1)
    return hNew


def syst_min_and_max_env_hist(h, proj_ax, syst_ax, indices, no_flow=[]):
    logger.debug(f"Taking the envelope of variation axis {syst_ax}, indices {indices}")
    hup = syst_min_or_max_env_hist(
        h, proj_ax, syst_ax, indices, no_flow=no_flow, do_min=False
    )
    hdown = syst_min_or_max_env_hist(
        h, proj_ax, syst_ax, indices, no_flow=no_flow, do_min=True
    )

    down_up_axis = hist.axis.Regular(
        2, -2.0, 2.0, underflow=False, overflow=False, name="downUpVar"
    )

    hnew = hist.Hist(*hup.axes, down_up_axis, storage=hup.storage_type())
    hnew[..., 0] = hdown.view(flow=True)
    hnew[..., 1] = hup.view(flow=True)
    return hnew


def syst_min_or_max_env_hist(h, proj_ax, syst_ax, indices, no_flow=[], do_min=True):
    if syst_ax not in h.axes.name:
        logger.warning(
            f"Did not find syst axis {syst_ax} in histogram. Returning nominal!"
        )
        return h

    # Keep the order of the hist
    proj_ax = [ax for ax in h.axes.name if ax in proj_ax]
    systax_idx = h.axes.name.index(syst_ax)
    if systax_idx != h.ndim - 1:
        raise ValueError("Required to have the syst axis at index -1")

    if len(indices) < 2:
        logger.warning(
            f"Requires at least two histograms for envelope. Returning nominal!"
        )
        return h

    if type(indices[0]) == str:
        if all(x.isdigit() for x in indices):
            indices == [int(x) for x in indices]
        else:
            indices = h.axes[syst_ax].index(indices)

    if max(indices) > h.axes[syst_ax].size:
        logger.warning(
            f"Range of indices exceeds length of syst axis '{syst_ax}.' Returning nominal!"
        )
        return h

    if syst_ax in proj_ax:
        proj_ax.pop(proj_ax.index(syst_ax))

    hvar = projectNoFlow(h, (*proj_ax, syst_ax), exclude=no_flow)

    view = np.take(hvar.view(flow=True), indices, axis=-1)
    fullview = np.take(h.view(flow=True), indices, axis=-1)

    # Move project axis to second to last position so the broadcasting works
    # NOTE: Be careful that you keep track of the actual order, keeping in mind that things are moving
    names = list(h.axes.name)
    initial_order = []
    for ax in proj_ax:
        idx = names.index(ax)
        initial_order.append(idx)
        # Inserts at second to last position
        names.insert(-1, names.pop(idx))
        # Moves axis to second to last position
        fullview = np.moveaxis(fullview, idx, -2)

    op = np.argmin if do_min else np.argmax
    # Index of min/max values considering only the eventual projection
    idx = op(view.value if hasattr(view, "value") else view, axis=-1)
    opview = fullview[(*np.indices(fullview.shape[:-1]), idx)]

    hnew = h[{syst_ax: 0}]
    # Now that the syst ax has been collapsed, project axes will be at last position
    # Move the axes back to where they belong
    names = list(h.axes.name)
    for idx in reversed(initial_order):
        opview = np.moveaxis(opview, -1, idx)

    hnew[...] = opview
    return hnew


def combineUpDownVarHists(down_hist, up_hist):
    if up_hist.axes != down_hist.axes:
        raise RuntimeError(
            "input up and down histograms have different axes, can't combine"
        )
    else:
        down_up_axis = hist.axis.Regular(
            2, -2.0, 2.0, underflow=False, overflow=False, name="downUpVar"
        )

        hnew = hist.Hist(
            *up_hist.axes, down_up_axis, storage=up_hist.storage_type()
        )
        hnew.view(flow=True)[...] = np.stack(
            (down_hist.view(flow=True), up_hist.view(flow=True)), axis=-1
        )
        return hnew


def smoothTowardsOne(h):
    if h.storage_type == hist.storage.Double:
        logger.warning(
            "Tried to smoothTowardsOne but histogram has no variances. Proceed without doing anything!"
        )
        return h

    vals = h.values(flow=True)
    vars = h.variances(flow=True)
    relErr = np.minimum(1.0, relVariance(vals, vars, fillOnes=True))
    newvals = (1.0 - relErr) * vals + relErr
    hnew = hist.Hist(*h.axes, storage=hist.storage.Weight())
    hnew.values(flow=True)[...] = newvals
    hnew.variances(flow=True)[...] = vars
    return hnew


def set_flow(h, val="nearest"):
    # sets the values of the underflow and overflow bins to given value; if val='nearest' the values of the nearest bins are taken
    for axis in [a.name for a in h.axes if a.traits.overflow]:
        slices = [slice(None) if a != axis else -1 for a in h.axes.name]
        h.view(flow=True)[tuple(slices)] = (
            h[{axis: -1}].view(flow=True) if val == "nearest" else val
        )
    for axis in [a.name for a in h.axes if a.traits.underflow]:
        slices = [slice(None) if a != axis else 0 for a in h.axes.name]
        h.view(flow=True)[tuple(slices)] = (
            h[{axis: 0}].view(flow=True) if val == "nearest" else val
        )
    return h


# For converting the helicity scale hist to variations, keeping the gen axis to be fit
# If swap_axes = True, the new axis takes the place of the old gen axis in the ordering
def expand_hist_by_duplicate_axis(
    href, ref_ax_name, new_ax_name, swap_axes=False, put_trailing=False, flow=True
):
    if ref_ax_name not in href.axes.name:
        raise ValueError(f"Did not find axis {ref_ax_name} in hist!")

    axes = href.axes

    ref_ax_idx = axes.name.index(ref_ax_name)
    ref_ax = axes[ref_ax_name]
    new_ax = copy.deepcopy(ref_ax)
    new_ax._ax.metadata["name"] = new_ax_name

    if put_trailing:
        new_axes = [*axes, new_ax]
        new_axis_idx = -1
    else:
        new_axes = [new_ax, *axes]
        new_axis_idx = 0
    if swap_axes:
        new_axes[new_axis_idx], new_axes[ref_ax_idx + 1] = (
            new_axes[ref_ax_idx + 1],
            new_axes[new_axis_idx],
        )
    hnew = hist.Hist(*new_axes, storage=href.storage_type())

    # Temporarily move axis to duplicate first
    data = np.moveaxis(href.values(flow=flow), ref_ax_idx, 0)
    # Copy data from other axis along the diagnoal, off-diagonals will be zero
    exp_data = data * np.reshape(
        np.identity(data.shape[0]),
        (data.shape[0], data.shape[0], *(1 for i in data.shape[1:])),
    )
    # Move original and duplicated axes to desired positions
    exp_data = np.moveaxis(exp_data, 0, new_axis_idx)
    hnew.values(flow=flow)[...] = np.moveaxis(
        exp_data, 0, ref_ax_idx + (new_axis_idx != -1 and new_axis_idx < ref_ax_idx)
    )  # move back one further if new axis comes before old one
    if href.storage_type == hist.storage.Weight:
        var = np.moveaxis(href.variances(flow=flow), ref_ax_idx, 0)
        exp_var = var * np.reshape(
            np.identity(var.shape[0]),
            (var.shape[0], var.shape[0], *(1 for i in var.shape[1:])),
        )
        exp_var = np.moveaxis(exp_var, 0, new_axis_idx)
        hnew.variances(flow=flow)[...] = np.moveaxis(
            exp_var, 0, ref_ax_idx + (new_axis_idx != -1 and new_axis_idx < ref_ax_idx)
        )

    return hnew


def expand_hist_by_duplicate_axes(href, ref_ax_names, new_ax_names, **kwargs):
    for ax_name, new_ax_name in zip(ref_ax_names, new_ax_names):
        href = expand_hist_by_duplicate_axis(href, ax_name, new_ax_name, **kwargs)
    return href


def swap_histogram_bins(
    histo,
    axis1,
    axis1_bin1,
    axis1_bin2,
    axis2=None,
    axis2_slice=None,
    flow=True,
    axis1_replace=None,
):
    # swap content from axis1: axis1_bin1 with axis1: axis1_bin2
    # optionally for a subset of the histogram defined by axis2: axis2_slice
    # optionally the selected bin content can be replaced by axis1_replace (example use case: setting up and down variations to nominal)
    if axis2 is not None and axis2_slice is None:
        raise ValueError(
            f"Requested to flip bins for axis {axis2} but the corresponding slices 'axis2_slice' are not set"
        )
    if isinstance(axis2_slice, slice):
        # for some reason complex slicing didn't work, convert to bin number
        tmp_slice = []
        for x in ("start", "stop", "step"):
            s = getattr(axis2_slice, x)
            if isinstance(s, complex):
                tmp_slice.append(histo.axes[axis2].index(s.imag))
            else:
                tmp_slice.append(s)
        axis2_slice = slice(*tmp_slice)

    slices1 = []
    slices2 = []
    slicesR = []
    for a in histo.axes.name:
        if a == axis1:
            slices1.append(histo.axes[a].index(axis1_bin1))
            slices2.append(histo.axes[a].index(axis1_bin2))
            if axis1_replace:
                slicesR.append(histo.axes[a].index(axis1_replace))
        elif axis2 is not None and a == axis2:
            slices1.append(axis2_slice)
            slices2.append(axis2_slice)
            if axis1_replace:
                slicesR.append(axis2_slice)
        else:
            slices1.append(slice(None))
            slices2.append(slice(None))
            if axis1_replace:
                slicesR.append(slice(None))

    # swap bins in specified slices
    data = histo.view(flow=flow)
    new_histo = histo.copy()
    new_histo.view(flow=flow)[tuple(slices2)] = (
        data[tuple(slices1)] if axis1_replace is None else data[tuple(slicesR)]
    )
    new_histo.view(flow=flow)[tuple(slices1)] = (
        data[tuple(slices2)] if axis1_replace is None else data[tuple(slicesR)]
    )
    return new_histo


def rescaleBandVariation(histo, factor):

    if factor == 1.0:
        return histo
    else:
        upper_env = histo.values()[..., 1]
        lower_env = histo.values()[..., 0]

        var = np.abs(upper_env - lower_env) / 2
        centr = (upper_env + lower_env) / 2
        new_upper = factor * var + centr
        new_lower = -factor * var + centr

        # leave sigmaUL variation to 50%
        new_upper[..., 0] = 1.5 * np.ones((new_upper.shape[0], new_upper.shape[1]))
        new_lower[..., 0] = 0.5 * np.ones((new_lower.shape[0], new_lower.shape[1]))

        histo[...] = np.stack([new_lower, new_upper], axis=-1)
        return histo


def rssHists(
    h, syst_axis, scale=1.0, hnom=None, returnDiff=False, returnDiffSquare=False
):
    s = hist.tag.Slicer()

    if hnom is None:
        hnom = h[{syst_axis: 0}]

    hdiff = addHists(h, hnom, scale2=-1.0) * scale
    hss = multiplyHists(hdiff, hdiff)
    if returnDiffSquare:
        return hss[{syst_axis: s[0 : hist.overflow : hist.sum]}]

    hrss = sqrtHist(hss[{syst_axis: s[0 : hist.overflow : hist.sum]}])
    if returnDiff:
        return hrss
    hUp = addHists(hnom, hrss)
    hDown = addHists(hnom, hrss, scale2=-1.0)

    return hUp, hDown


def rssHistsMid(h, syst_axis, scale=1.0):
    s = hist.tag.Slicer()

    hnom = 0.5 * (
        h[{"downUpVar": -1j, syst_axis: 0}] + h[{"downUpVar": 1j, syst_axis: 0}]
    )
    hdiff = addHists(h, hnom, scale2=-1.0) * scale
    hss = multiplyHists(hdiff, hdiff)

    hrss = sqrtHist(hss[{syst_axis: s[0 : hist.overflow : hist.sum]}])
    hUp = addHists(hnom, hrss[{"downUpVar": 1j}])
    hDown = addHists(hnom, hrss[{"downUpVar": -1j}], scale2=-1.0)

    return hUp, hDown



import datetime
import json
import math
import pathlib
import shutil
import socket
import sys
import textwrap

import hist
import matplotlib as mpl
import matplotlib.pyplot as plt
import mplhep as hep
import numpy as np
from matplotlib import patches, ticker
from matplotlib.legend_handler import HandlerLine2D, HandlerPatch
from matplotlib.lines import Line2D
from matplotlib.patches import Polygon

# for setting number of decimal places on tick labels
from matplotlib.ticker import StrMethodFormatter

from wums import boostHistHelpers as hh
from wums import ioutils, logging

hep.style.use(hep.style.ROOT)

logger = logging.child_logger(__name__)


def cfgFigure(
    href, xlim=None, bin_density=300, width_scale=1, automatic_scale=True, base_size=8
):
    hax = href.axes[0]
    if not xlim:
        xlim = [hax.edges[0], hax.edges[-1]]
    if not automatic_scale:
        return plt.figure(figsize=(width_scale * base_size, base_size)), xlim
    xlim_range = float(xlim[1] - xlim[0])
    original_xrange = float(hax.edges[-1] - hax.edges[0])
    raw_width = (hax.size / float(bin_density)) * (xlim_range / original_xrange)
    width = math.ceil(raw_width)
    return (
        plt.figure(figsize=(width_scale * base_size * width, width_scale * base_size)),
        xlim,
    )


def figure(
    href,
    xlabel,
    ylabel,
    ylim=None,
    xlim=None,
    grid=False,
    plot_title=None,
    title_padding=0,
    bin_density=300,
    logy=False,
    logx=False,
    width_scale=1,
    height=8,
    automatic_scale=True,
):
    if isinstance(href, hist.Hist):
        fig, xlim = cfgFigure(href, xlim, bin_density, width_scale, automatic_scale)
    else:
        if automatic_scale:
            raw_width = len(href) / float(bin_density)
            width = math.ceil(raw_width)
        else:
            width = 1
        fig = plt.figure(figsize=(width_scale * height * width, height))

    ax1 = fig.add_subplot()

    ax1.set_xlabel(xlabel)
    if ylabel is not None:
        ax1.set_ylabel(ylabel, labelpad=40, va="center", multialignment="center")
    ax1.set_xlim(xlim)

    if ylim is not None:
        ax1.set_ylim(ylim)
    else:
        ax1.autoscale(axis="y")

    if logy:
        ax1.set_yscale("log")
    if logx:
        ax1.set_xscale("log")

    if grid:
        ax1.grid(which="both")
    if plot_title:
        ax1.set_title(plot_title, pad=title_padding)
    return fig, ax1


def figureWithRatio(
    href,
    xlabel,
    ylabel,
    ylim,
    rlabel,
    rrange,
    xlim=None,
    grid_on_main_plot=False,
    grid_on_ratio_plot=False,
    plot_title=None,
    title_padding=0,
    x_ticks_ndp=None,
    bin_density=300,
    logy=False,
    logx=False,
    width_scale=1,
    automatic_scale=True,
    base_size=8,
    only_ratio=False,
    subplotsizes=[4, 2],
):
    fig, xlim = cfgFigure(
        href, xlim, bin_density, width_scale, automatic_scale, base_size
    )

    ratio_axes = []

    if len(subplotsizes) == 2 and len(rrange) == 2 and isinstance(rlabel, str):
        rrange = [rrange]
        rlabel = [rlabel]

    if not only_ratio:
        ax1 = fig.add_subplot(sum(subplotsizes), 1, (1, subplotsizes[0]))
        ax1.set_xlabel(" ")
        ax1.set_ylabel(ylabel, labelpad=40, va="center", multialignment="center")
        ax1.set_xlim(xlim)

        if ylim:
            ax1.set_ylim(ylim)
        else:
            ax1.autoscale(axis="y")

        if logy:
            ax1.set_yscale("log")
        if grid_on_main_plot:
            ax1.grid(which="both")
        if plot_title:
            ax1.set_title(plot_title, pad=title_padding)

    for i, (ax, rr, rl) in enumerate(zip(subplotsizes[1:], rrange, rlabel)):
        xax = fig.add_subplot(
            sum(subplotsizes),
            1,
            (sum(subplotsizes[: i + 1]) + 1, sum(subplotsizes[: i + 2])),
        )
        xax.set_xlabel(" ")
        xax.set_ylabel(rl, labelpad=40, va="center", multialignment="center")
        xax.set_ylim(rr)
        ratio_axes.append(xax)

    for ax in ratio_axes:
        ax.set_xlim(xlim)

        if x_ticks_ndp:
            ax.xaxis.set_major_formatter(
                StrMethodFormatter("{x:." + str(x_ticks_ndp) + "f}")
            )
        if logx:
            ax.set_xscale("log")

        if grid_on_ratio_plot:
            ax.grid(which="both")

    ratio_axes[-1].set_xlabel(xlabel)

    if logx and not only_ratio:
        ax1.set_xscale("log")

    if not only_ratio:
        return fig, ax1, ratio_axes
    else:
        return fig, xax


class StackedLineHandler:
    def legend_artist(
        self, legend, orig_handle, fontsize, handlebox, linewidth_scale=1.0
    ):
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width, height = handlebox.width, handlebox.height
        line1 = Line2D(
            [x0, x0 + width],
            [y0 + height * 0.75, y0 + height * 0.75],
            color=orig_handle.get_color(),
            lw=linewidth_scale * orig_handle.get_linewidth(),
            linestyle="-",
        )
        line2 = Line2D(
            [x0, x0 + width],
            [y0 + height * 0.25, y0 + height * 0.25],
            color=orig_handle.get_color(),
            lw=linewidth_scale * orig_handle.get_linewidth(),
            linestyle="--",
        )
        handlebox.add_artist(line1)
        handlebox.add_artist(line2)
        return [line1, line2]


class StackFilledHandler:
    def legend_artist(
        self, legend, orig_handle, fontsize, handlebox, linewidth_scale=1.0
    ):
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width, height = handlebox.width, handlebox.height
        line0 = Line2D(
            [x0, x0 + width],
            [y0 + height * 0.5, y0 + height * 0.5],
            color=orig_handle.get_edgecolor(),
            lw=linewidth_scale * orig_handle.get_linewidth(),
            linestyle="-",
        )
        line1 = Line2D(
            [x0, x0 + width],
            [y0 + height, y0 + height],
            color=orig_handle.get_edgecolor(),
            lw=linewidth_scale * orig_handle.get_linewidth(),
            linestyle=orig_handle.get_linestyle(),
        )
        line2 = Line2D(
            [x0, x0 + width],
            [y0, y0],
            color=orig_handle.get_edgecolor(),
            lw=linewidth_scale * orig_handle.get_linewidth(),
            linestyle=orig_handle.get_linestyle(),
        )
        # Create the filled area between the lines using a polygon
        fill_coords = [
            [x0, y0],
            [x0 + width, y0],
            [x0 + width, y0 + height],
            [x0, y0 + height],
        ]
        fill = Polygon(fill_coords, color=orig_handle.get_facecolor(), alpha=0.3)

        handlebox.add_artist(fill)
        handlebox.add_artist(line0)
        handlebox.add_artist(line1)
        handlebox.add_artist(line2)
        return [line0, line1, line2, fill]


class BandFilledHandler:
    def legend_artist(
        self, legend, orig_handle, fontsize, handlebox, linewidth_scale=1.0
    ):
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width, height = handlebox.width, handlebox.height
        line1 = Line2D(
            [x0, x0 + width],
            [y0 + height, y0 + height],
            color=orig_handle.get_edgecolor(),
            lw=linewidth_scale * orig_handle.get_linewidth(),
            linestyle=orig_handle.get_linestyle(),
        )
        line2 = Line2D(
            [x0, x0 + width],
            [y0, y0],
            color=orig_handle.get_edgecolor(),
            lw=linewidth_scale * orig_handle.get_linewidth(),
            linestyle=orig_handle.get_linestyle(),
        )
        # Create the filled area between the lines using a polygon
        fill_coords = [
            [x0, y0],
            [x0 + width, y0],
            [x0 + width, y0 + height],
            [x0, y0 + height],
        ]
        fill = Polygon(fill_coords, color=orig_handle.get_facecolor(), alpha=0.3)

        handlebox.add_artist(fill)
        handlebox.add_artist(line1)
        handlebox.add_artist(line2)
        return [line1, line2, fill]


class DoubleBandHandler(HandlerPatch):
    def create_artists(
        self,
        legend,
        orig_handle,
        xdescent,
        ydescent,
        width,
        height,
        fontsize,
        trans,
        linewidth_scale=1.0,
    ):
        # Create the outer polygon
        polygon_outer = Polygon(
            [
                [xdescent, ydescent],
                [xdescent + width, ydescent],
                [xdescent + width, ydescent + height],
                [xdescent, ydescent + height],
            ],
            color=orig_handle.get_facecolor(),
            alpha=orig_handle.get_alpha(),
        )
        # Create the inner polygon (narrower)
        margin = 0.25 * width
        polygon_inner = Polygon(
            [
                [xdescent + margin, ydescent],
                [xdescent + width - margin, ydescent],
                [xdescent + width - margin, ydescent + height],
                [xdescent + margin, ydescent + height],
            ],
            color=orig_handle.get_facecolor(),
            alpha=orig_handle.get_alpha(),
        )
        line = Line2D(
            [xdescent + width / 2, xdescent + width / 2],
            [ydescent, ydescent + height],
            color="black",
            lw=linewidth_scale * orig_handle.get_linewidth(),
            linestyle=orig_handle.get_linestyle(),
        )

        # Set the transformations
        polygon_outer.set_transform(trans)
        polygon_inner.set_transform(trans)
        line.set_transform(trans)

        return [polygon_outer, polygon_inner, line]


class TripleBandHandler(HandlerPatch):
    def create_artists(
        self,
        legend,
        orig_handle,
        xdescent,
        ydescent,
        width,
        height,
        fontsize,
        trans,
        linewidth_scale=1.0,
    ):
        # Create the outer polygon
        polygon_outer = Polygon(
            [
                [xdescent, ydescent],
                [xdescent + width, ydescent],
                [xdescent + width, ydescent + height],
                [xdescent, ydescent + height],
            ],
            color=orig_handle.get_facecolor(),
            alpha=orig_handle.get_alpha(),
        )
        # Create the inner polygon (narrower)
        margin = 0.15 * width
        polygon_inner = Polygon(
            [
                [xdescent + margin, ydescent],
                [xdescent + width - margin, ydescent],
                [xdescent + width - margin, ydescent + height],
                [xdescent + margin, ydescent + height],
            ],
            color=orig_handle.get_facecolor(),
            alpha=orig_handle.get_alpha(),
        )
        # Create the inner polygon (narrower)
        margin = 0.3 * width
        polygon_inner2 = Polygon(
            [
                [xdescent + margin, ydescent],
                [xdescent + width - margin, ydescent],
                [xdescent + width - margin, ydescent + height],
                [xdescent + margin, ydescent + height],
            ],
            color=orig_handle.get_facecolor(),
            alpha=orig_handle.get_alpha(),
        )
        line = Line2D(
            [xdescent + width / 2, xdescent + width / 2],
            [ydescent, ydescent + height],
            color="black",
            lw=linewidth_scale * orig_handle.get_linewidth(),
            linestyle=orig_handle.get_linestyle(),
        )

        # Set the transformations
        polygon_outer.set_transform(trans)
        polygon_inner.set_transform(trans)
        polygon_inner2.set_transform(trans)
        line.set_transform(trans)

        return [polygon_outer, polygon_inner, polygon_inner2, line]


def get_custom_handler_map(keys):
    if len(keys) == 0:
        return None
    handler_map = {}
    for key in keys:
        if key == "stacked":
            handler_map[Line2D] = StackedLineHandler()
        elif key == "stackfilled":
            handler_map[Polygon] = StackFilledHandler()
        elif key == "bandfilled":
            handler_map[Polygon] = BandFilledHandler()
        elif key == "verticleline":
            handler_map[Line2D] = HandlerLine2D(update_func=update_prop)
        elif key == "doubleband":
            handler_map[Polygon] = DoubleBandHandler()
        elif key == "tripleband":
            handler_map[Polygon] = TripleBandHandler()
    return handler_map


# https://stackoverflow.com/questions/53122592/legend-with-vertical-line
def update_prop(handle, orig):
    handle.update_from(orig)
    x, y = handle.get_data()
    handle.set_data([np.mean(x)] * 2, [0, 2 * y[0]])


def padding(ncols, labels, handles, loc="auto"):
    if len(labels) % ncols:
        # if not all legend places are filled, add empty legend entries towards the center of the figure
        rest = ncols - len(labels) % ncols
        nrows = int(np.ceil(len(labels) / ncols))
        for i in range(rest):
            if loc == "upper right":
                idx = (i + 1) * nrows - 1
            elif loc == "lower right":
                idx = i * nrows
            elif loc == "upper left":
                idx = len(labels) - i * nrows
            elif loc == "lower left":
                idx = len(labels) - (i + 1) * nrows + 1
            else:
                raise ValueError(f"Invalid option {loc} for legend padding location")
            handles.insert(idx, patches.Patch(color="none", label=" "))
            labels.insert(idx, " ")
    return labels, handles


def addLegend(
    ax,
    ncols=2,
    extra_text=None,
    extra_text_loc=None,
    text_size=None,
    loc="upper right",
    bbox_to_anchor=None,
    extra_handles=[],
    extra_labels=[],
    extra_entries_first=True,
    custom_handlers=[],
    markerfirst=True,
    reverse=True,
    labelcolor=None,
    padding_loc="auto",
):
    handles, labels = ax.get_legend_handles_labels()
    if extra_entries_first:
        handles.extend(extra_handles)
        labels.extend(extra_labels)
    else:
        handles = [*extra_handles, *handles]
        labels = [*extra_labels, *labels]

    if len(labels) == 0:
        return

    if padding_loc == "auto":
        legend_to_padding_loc = {
            "upper right": "lower left",
            "1": "lower left",
            "upper left": "lower right",
            "2": "lower right",
            "lower left": "upper right",
            "3": "upper right",
            "lower right": "upper left",
            "4": "upper left",
        }
        padding_loc = legend_to_padding_loc.get(str(loc), "lower right")

    labels, handles = padding(ncols, labels, handles, padding_loc)

    text_size = get_textsize(ax, text_size)
    handler_map = get_custom_handler_map(custom_handlers)
    leg = ax.legend(
        handles=handles,
        labels=labels,
        prop={"size": text_size},
        ncol=ncols,
        loc=loc,
        handler_map=handler_map,
        reverse=reverse,
        bbox_to_anchor=bbox_to_anchor,
        markerfirst=markerfirst,
        labelcolor=labelcolor,
    )

    if extra_text is not None:
        if extra_text_loc is None:
            # Add text to the left of the legend
            # Get the bounding box of the legend
            bbox = leg.get_window_extent()

            # Convert the bbox to display coordinates (relative to the figure)
            bbox_transform = plt.gcf().transFigure.inverted()
            bbox_disp = bbox_transform.transform(bbox)

            # Adjust the x position by moving it to the left
            extra_text_loc = bbox_disp[0, 0] - 0.25, bbox_disp[1, 1] - 0.01

            transform = plt.gcf().transFigure
        else:
            transform = None

        wrap_text(
            extra_text,
            ax,
            *extra_text_loc,
            text_size=text_size,
            ha="left",
            va="top",
            transform=transform,
        )


def get_textsize(ax, text_size):
    if text_size == "large" or text_size is None:
        # legend size same as axis label size
        return ax.yaxis.label.get_size()
    elif text_size == "small":
        # legend size same as axis ticklabel size (numbers)
        return ax.yaxis.get_ticklabels()[0].get_fontsize()
    elif text_size == "verysmall":
        # legend size same as axis ticklabel size (numbers)
        return ax.yaxis.get_ticklabels()[0].get_fontsize() * 0.6
    else:
        return int(text_size)


def wrap_text(
    text,
    ax,
    lower_x,
    y,
    upper_x=None,
    text_size=None,
    transform=None,
    ha=None,
    va="center",
):
    # wrap text within lower_x and upper_x,
    #  if text is already given as pieces in a list, use these pieces,
    #  otherwise calculate the pieces automatically
    text_size = get_textsize(ax, text_size)

    if isinstance(text, str):
        # Get the width of the text in data coordinates
        bbox = ax.get_window_extent().transformed(ax.transData.inverted())
        width_data = upper_x - lower_x
        width_display = bbox.width * (
            width_data / (ax.get_xlim()[1] - ax.get_xlim()[0])
        )
        # Estimate the number of characters that fit in this width
        # This is an approximation and may need adjustment
        char_width = text_size * 0.2  # Approximate width of a character in inches
        max_chars = int(width_display / char_width)
        wrapped_text = "\n".join(textwrap.wrap(text, width=max_chars))
    else:
        wrapped_text = "\n".join(text)

    if ha is not None:
        x = lower_x
    elif upper_x is not None:
        x = (lower_x + upper_x) / 2
        ha = "center"
    else:
        x = lower_x
        ha = "left"
    ax.text(
        x,
        y,
        wrapped_text,
        ha=ha,
        va=va,
        transform=transform if transform is not None else ax.transAxes,
        fontsize=text_size,
        wrap=True,
    )


def add_cms_decor(
    ax, label=None, lumi=None, loc=2, data=True, text_size=None, no_energy=False
):
    text_size = get_textsize(ax, text_size)
    if no_energy:
        hep.cms.text(ax=ax, text=label, loc=loc, fontsize=text_size)
    else:
        hep.cms.label(
            ax=ax,
            lumi=lumi,
            lumi_format="{0:.3g}",
            fontsize=text_size,
            label=label,
            data=data,
            loc=loc,
        )


def makeStackPlotWithRatio(
    histInfo,
    stackedProcs,
    histName="nominal",
    unstacked=None,
    fitresult=None,
    prefit=False,
    xlabel="",
    ylabel=None,
    rlabel="Data/Pred.",
    rrange=[0.9, 1.1],
    ylim=None,
    xlim=None,
    nlegcols=2,
    binwnorm=None,
    select={},
    action=(lambda x: x),
    extra_text=None,
    extra_text_loc=(0.8, 0.7),
    grid=False,
    plot_title=None,
    title_padding=0,
    yscale=None,
    logy=False,
    logx=False,
    fill_between=False,
    ratio_to_data=False,
    baseline=True,
    legtext_size=20,
    cms_decor="Preliminary",
    lumi=16.8,
    no_fill=False,
    no_stack=False,
    no_ratio=False,
    density=False,
    flow="none",
    bin_density=300,
    unstacked_linestyles=[],
    double_lines=False,
    ratio_error=True,
    normalize_to_data=False,
    cutoff=1e-6,
    noSci=False,
    logoPos=2,
    width_scale=1.0,
    linewidth=2,
    alpha=0.7,
    legPos="upper right",
    leg_padding="auto",
    lowerLegCols=2,
    lowerLegPos="upper right",
    lower_panel_variations=0,
    lower_leg_padding="auto",
    scaleRatioUnstacked=[],
    subplotsizes=[4, 2],
):
    add_ratio = not (no_stack or no_ratio)
    if ylabel is None:
        ylabel = "Events/bin" if not density else "density"

    colors = [histInfo[k].color for k in stackedProcs if histInfo[k].hists[histName]]
    labels = [histInfo[k].label for k in stackedProcs if histInfo[k].hists[histName]]

    to_read = stackedProcs[:]
    if "Data" in histInfo:
        to_read.append("Data")

    stack = []
    data_hist = None
    for k in to_read:
        if histName not in histInfo[k].hists or not histInfo[k].hists[histName]:
            logger.warning(f"Failed to find hist {histName} for proc {k}")
            continue
        h = action(histInfo[k].hists[histName])[select]

        # Use this if the hist has been rebinned for combine
        if xlim:
            h = h[complex(0, xlim[0]) : complex(0, xlim[1])]

        # If plotting from combine, apply the action to the underlying hist.
        # Don't do this for the generic case, as it screws up the ability to make multiple plots
        if fitresult:
            histInfo[k].hists[histName] = h

        if k != "Data":
            stack.append(h)
        else:
            data_hist = h

    if add_ratio:
        fig, ax1, ratio_axes = figureWithRatio(
            stack[0],
            xlabel,
            ylabel,
            ylim,
            rlabel,
            rrange,
            xlim=xlim,
            logy=logy,
            logx=logx,
            grid_on_ratio_plot=grid,
            plot_title=plot_title,
            title_padding=title_padding,
            bin_density=bin_density,
            width_scale=width_scale,
            subplotsizes=subplotsizes,
        )
        ax2 = ratio_axes[-1]
    else:
        fig, ax1 = figure(
            stack[0],
            xlabel,
            ylabel,
            ylim,
            xlim=xlim,
            logy=logy,
            logx=logx,
            plot_title=plot_title,
            title_padding=title_padding,
            bin_density=bin_density,
            width_scale=width_scale,
        )
        ratio_axes = None
        ax2 = None

    if fitresult:
        import uproot

        combine_result = uproot.open(fitresult)

        fittype = "prefit" if prefit else "postfit"

        # set histograms to prefit/postfit values
        for p in to_read:

            hname = f"expproc_{p}_{fittype}" if p != "Data" else "obs"
            vals = combine_result[hname].to_hist().values()
            if len(histInfo[p].hists[histName].values()) != len(vals):
                raise ValueError(
                    f"The size of the combine histogram ({(vals.shape)}) is not consistent with the xlim or input hist ({histInfo[p].hists[histName].shape})"
                )

            histInfo[p].hists[histName].values()[...] = vals
            if p == "Data":
                histInfo[p].hists[histName].variances()[...] = vals

        # for postfit uncertaity bands
        axis = histInfo[to_read[0]].hists[histName].axes[0].edges

        # need to divide by bin width
        binwidth = axis[1:] - axis[:-1]
        hexp = combine_result[f"expfull_{fittype}"].to_hist()
        if hexp.storage_type != hist.storage.Weight:
            raise ValueError(
                f"Did not find uncertainties in {fittype} hist. Make sure you run combinetf with --computeHistErrors!"
            )
        nom = hexp.values() / binwidth
        std = np.sqrt(hexp.variances()) / binwidth

        hatchstyle = "///"
        ax1.fill_between(
            axis,
            np.append(nom + std, (nom + std)[-1]),
            np.append(nom - std, (nom - std)[-1]),
            step="post",
            facecolor="none",
            zorder=2,
            hatch=hatchstyle,
            edgecolor="k",
            linewidth=0.0,
            label="Uncertainty",
        )

        if add_ratio:
            ax2.fill_between(
                axis,
                np.append((nom + std) / nom, ((nom + std) / nom)[-1]),
                np.append((nom - std) / nom, ((nom - std) / nom)[-1]),
                step="post",
                facecolor="none",
                zorder=2,
                hatch=hatchstyle,
                edgecolor="k",
                linewidth=0.0,
            )

    opts = dict(stack=not no_stack, flow=flow)
    optsr = opts.copy()  # no binwnorm for ratio axis
    optsr["density"] = density
    if density:
        opts["density"] = True
    else:
        opts["binwnorm"] = binwnorm

    if type(unstacked) == str:
        unstacked = unstacked.split(",")

    scale = 1.0
    if normalize_to_data:
        if "Data" not in histInfo:
            raise ValueError("Can't normalize to data without a data histogram!")

        vals = [
            x.value if hasattr(x, "value") else x
            for x in (data_hist.sum(), hh.sumHists(stack).sum())
        ]
        varis = [
            x.variance if hasattr(x, "variance") else x ** 0.5
            for x in (data_hist.sum(), hh.sumHists(stack).sum())
        ]
        scale = vals[0] / vals[1]
        unc = scale * (varis[0] / vals[0] ** 2 + varis[1] / vals[1] ** 2) ** 0.5
        ndigits = -math.floor(math.log10(abs(unc))) + 1
        logger.info(
            f"Rescaling all processes by {round(scale,ndigits)} +/- {round(unc,ndigits)} to match data norm"
        )
        stack = [s * scale for s in stack]

    hep.histplot(
        stack,
        histtype="fill" if not no_fill else "step",
        color=colors,
        label=labels,
        ax=ax1,
        zorder=1,
        **opts,
    )

    if "Data" in histInfo and ratio_to_data and add_ratio:
        hep.histplot(
            hh.divideHists(
                hh.sumHists(stack), data_hist, cutoff=cutoff, by_ax_name=False
            ),
            histtype="step",
            color=histInfo[stackedProcs[-1]].color,
            label=histInfo[stackedProcs[-1]].label,
            yerr=False,
            ax=ax2,
            zorder=3,
            **optsr,
        )

    extra_handles = []
    extra_labels = []
    if unstacked:
        linestyles = ["solid"] * len(unstacked)
        data_idx = -1
        if "Data" in unstacked:
            data_idx = unstacked.index("Data")
            linestyles[data_idx] = "None"
        linestyles = np.array(linestyles, dtype=object)
        logger.debug("Number of linestyles", len(linestyles))
        logger.debug("Length of unstacked", len(unstacked))
        if unstacked_linestyles:
            linestyles[data_idx + 1 : data_idx + 1 + len(unstacked_linestyles)] = (
                unstacked_linestyles
            )
        elif double_lines:
            linestyles[data_idx + 1 :: 2] = ["solid"] * len(
                linestyles[data_idx + 1 :: 2]
            )
            linestyles[data_idx + 2 :: 2] = ["dashed"] * len(
                linestyles[data_idx + 2 :: 2]
            )

        ratio_ref = data_hist if ratio_to_data else hh.sumHists(stack)
        if baseline and add_ratio:
            hep.histplot(
                hh.divideHists(
                    ratio_ref,
                    ratio_ref,
                    cutoff=cutoff,
                    rel_unc=True,
                    flow=False,
                    by_ax_name=False,
                ),
                histtype="step",
                color="grey",
                alpha=0.5,
                yerr=(
                    ratio_error
                    if ratio_ref.storage_type == hist.storage.Weight
                    else False
                ),
                ax=ax2,
                linewidth=linewidth,
                **optsr,
            )

        if fill_between and add_ratio:
            fill_procs = [x for x in unstacked if x != "Data"]
            if fill_between < 0:
                fill_between = len(fill_procs) + 1
            logger.debug(f"Filling first {fill_between}")
            for up, down in zip(
                fill_procs[:fill_between:2], fill_procs[1:fill_between:2]
            ):
                unstack_up = action(histInfo[up].hists[histName]) * scale
                unstack_down = action(histInfo[down].hists[histName]) * scale
                unstack_upr = hh.divideHists(
                    unstack_up, ratio_ref, 1e-6, flow=False, by_ax_name=False
                ).values()
                unstack_downr = hh.divideHists(
                    unstack_down, ratio_ref, 1e-6, flow=False, by_ax_name=False
                ).values()
                ax2.fill_between(
                    unstack_up.axes[0].edges,
                    np.insert(unstack_upr, 0, unstack_upr[0]),
                    np.insert(unstack_downr, 0, unstack_downr[0]),
                    step="pre",
                    color=histInfo[up].color,
                    alpha=0.5,
                )

        for i, (proc, style) in enumerate(zip(unstacked, linestyles)):
            unstack = histInfo[proc].hists[histName]
            if not fitresult or proc not in to_read:
                unstack = action(unstack)[select]
            if proc != "Data":
                unstack = unstack * scale
            if len(scaleRatioUnstacked) > i:
                hdiff = hh.addHists(unstack, ratio_ref, scale2=-1)
                hdiff = hh.scaleHist(hdiff, scaleRatioUnstacked[i])
                unstack = hh.addHists(hdiff, ratio_ref)

            if i >= lower_panel_variations or proc == "Data":
                # unstacked that are filled between are only plot in the lower panel
                hep.histplot(
                    unstack,
                    yerr=True if style == "None" else False,
                    histtype="errorbar" if style == "None" else "step",
                    color=histInfo[proc].color,
                    label=histInfo[proc].label,
                    ax=ax1,
                    alpha=alpha if style != "None" else 1.0,
                    linestyle=style,
                    linewidth=linewidth,
                    **opts,
                )
            elif histInfo[proc].label and histInfo[proc].label not in extra_labels:
                if fill_between is not None and i < fill_between:
                    extra_handles.append(
                        Polygon(
                            [[0, 0], [0, 0], [0, 0], [0, 0]],
                            color=histInfo[proc].color,
                            linestyle=style,
                            linewidth=linewidth,
                            alpha=alpha,
                        )
                    )
                else:
                    extra_handles.append(
                        Line2D(
                            [0],
                            [0],
                            color=histInfo[proc].color,
                            linestyle=style,
                            linewidth=linewidth,
                        )
                    )

                extra_labels.append(histInfo[proc].label)
            if ratio_to_data and proc == "Data" or not add_ratio:
                continue
            stack_ratio = hh.divideHists(
                unstack,
                ratio_ref,
                cutoff=cutoff,
                rel_unc=True,
                flow=False,
                by_ax_name=False,
            )
            hep.histplot(
                stack_ratio,
                histtype="errorbar" if style == "None" else "step",
                color=histInfo[proc].color,
                yerr=(
                    True
                    if (
                        style == "None"
                        and stack_ratio.storage_type == hist.storage.Weight
                    )
                    else False
                ),
                linewidth=linewidth,
                linestyle=style,
                ax=ax2,
                **optsr,
            )

    addLegend(
        ax1,
        nlegcols,
        loc=legPos,
        extra_text=extra_text,
        extra_text_loc=extra_text_loc,
        text_size=legtext_size,
        padding_loc=leg_padding,
    )
    if add_ratio:
        addLegend(
            ax2,
            lowerLegCols,
            text_size=legtext_size,
            loc=lowerLegPos,
            extra_handles=extra_handles,
            extra_labels=extra_labels,
            custom_handlers=(
                ["bandfilled"]
                if fill_between is not None
                else ["stacked"] if double_lines else []
            ),
            padding_loc=lower_leg_padding,
        )

    fix_axes(ax1, ratio_axes, fig, yscale=yscale, logy=logy, noSci=noSci)

    lumi = float(f"{lumi:.3g}") if not density else None
    if cms_decor:
        add_cms_decor(ax1, cms_decor, data="Data" in histInfo, lumi=lumi, loc=logoPos)

    return fig


def makePlotWithRatioToRef(
    hists,
    labels,
    colors,
    hists_ratio=None,
    midratio_idxs=None,
    linestyles=[],
    xlabel="",
    ylabel="Events/bin",
    rlabel=["x/nominal"],
    rrange=[[0.9, 1.1]],
    ylim=None,
    xlim=None,
    nlegcols=2,
    lowerLegPos="upper right",
    lowerLegCols=2,
    binwnorm=None,
    alpha=1.0,
    baseline=True,
    dataIdx=None,
    ratio_to_data=False,
    autorrange=None,
    grid=False,
    extra_text=None,
    extra_text_loc=(0.8, 0.7),
    yerr=False,
    legtext_size=20,
    plot_title=None,
    x_ticks_ndp=None,
    bin_density=300,
    yscale=None,
    logoPos=2,
    logy=False,
    logx=False,
    fill_between=0,
    title_padding=0,
    cms_label=None,
    cutoff=1e-6,
    only_ratio=False,
    width_scale=1,
    automatic_scale=True,
    base_size=8,
    linewidth=2,
    leg_padding="auto",
    lower_leg_padding="auto",
    subplotsizes=[4, 2],
    no_sci=False,
    lumi=None,
    center_rlabels=False,
    swap_ratio_panels=False,
    select={},
):
    if select:
        hists = [h[select] for h in hists]

    if hists_ratio is None:
        hists_ratio = hists
    elif select is not None:
        hists_ratio = [h[select] for h in hists_ratio]

    if len(hists_ratio) != len(labels) or len(hists_ratio) != len(colors):
        raise ValueError(
            f"Number of hists ({len(hists_ratio)}), colors ({len(colors)}), and labels ({len(labels)}) must agree!"
        )
    ratio_hists = [
        hh.divideHists(
            h,
            hists[dataIdx if ratio_to_data else 0],
            cutoff=cutoff,
            flow=False,
            rel_unc=True,
            by_ax_name=False,
        )
        for h in hists_ratio[not baseline :]
    ]

    ratio_axes_idx = -1
    midratio_axes_idx = 0
    midratio_hists = None
    if len(subplotsizes) == 3:
        if midratio_idxs is None:
            raise ValueError("Found two ratio axes but no midratio hist indices!")

        midratio_hists = [
            hh.divideHists(
                hists_ratio[i],
                hists_ratio[midratio_idxs[0]],
                cutoff=cutoff,
                flow=False,
                rel_unc=True,
                by_ax_name=False,
            )
            for i in midratio_idxs
        ]

        if swap_ratio_panels:
            rlabel = rlabel[::-1]
            rrange = rrange[::-1]
            ratio_axes_idx = 0
            midratio_axes_idx = -1

    if not only_ratio:
        fig, ax1, ratio_axes = figureWithRatio(
            hists[0],
            xlabel,
            ylabel,
            ylim,
            rlabel,
            rrange,
            xlim=xlim,
            grid_on_ratio_plot=grid,
            plot_title=plot_title,
            title_padding=title_padding,
            bin_density=bin_density,
            logy=logy,
            logx=logx,
            only_ratio=only_ratio,
            width_scale=width_scale,
            subplotsizes=subplotsizes,
            automatic_scale=automatic_scale,
            base_size=base_size,
        )
        ax2 = ratio_axes[ratio_axes_idx]
    else:
        fig, ax2 = figureWithRatio(
            hists[0],
            xlabel,
            ylabel,
            ylim,
            rlabel,
            rrange,
            xlim=xlim,
            grid_on_ratio_plot=grid,
            plot_title=plot_title,
            title_padding=title_padding,
            bin_density=bin_density,
            logy=logy,
            logx=logx,
            only_ratio=only_ratio,
            width_scale=width_scale,
            automatic_scale=automatic_scale,
            base_size=base_size,
        )

    linestyles = linestyles + ["solid"] * (len(hists_ratio) - len(linestyles))

    exclude_data = lambda x: [j for i, j in enumerate(x) if i != dataIdx]

    if dataIdx is not None:
        hep.histplot(
            hists[dataIdx],
            histtype="errorbar",
            color=colors[dataIdx],
            label=labels[dataIdx],
            stack=False,
            ax=ax1,
            binwnorm=binwnorm,
            alpha=alpha,
            flow="none",
            zorder=4,
        )

    hists_noData = exclude_data(hists)
    if not only_ratio:
        hep.histplot(
            hists_noData,
            histtype="step",
            color=exclude_data(colors)[: len(hists_noData)],
            label=exclude_data(labels)[: len(hists_noData)],
            linestyle=exclude_data(linestyles)[: len(hists_noData)],
            linewidth=linewidth,
            stack=False,
            ax=ax1,
            yerr=yerr,
            binwnorm=binwnorm,
            alpha=alpha,
            flow="none",
            zorder=3,
        )

    if len(ratio_hists) > 1:
        plotRatio(
            ax2,
            ratio_hists=ratio_hists,
            labels=labels,
            colors=colors,
            linestyles=linestyles,
            linewidth=linewidth,
            lowerLegPos=lowerLegPos,
            lowerLegCols=lowerLegCols,
            legtext_size=legtext_size,
            lower_leg_padding=lower_leg_padding,
            alpha=alpha,
            yerr=False,
            fill_between=fill_between,
            dataIdx=dataIdx,
            baseline=baseline,
            add_legend=not only_ratio,
        )
        if midratio_hists:
            plotRatio(
                ratio_axes[midratio_axes_idx],
                ratio_hists=midratio_hists,
                labels=[labels[i] for i in midratio_idxs],
                colors=[colors[i] for i in midratio_idxs],
                linestyles=[linestyles[i] for i in midratio_idxs],
                linewidth=linewidth,
                lowerLegPos=lowerLegCols,
                lowerLegCols=lowerLegCols,
                legtext_size=legtext_size,
                lower_leg_padding=lower_leg_padding,
                alpha=alpha,
                yerr=False,
                fill_between=fill_between,
                dataIdx=(
                    midratio_idxs.index(dataIdx) if dataIdx in midratio_idxs else None
                ),
                baseline=baseline,
                add_legend=False,
            )

    if not only_ratio:
        addLegend(
            ax1,
            nlegcols,
            extra_text=extra_text,
            extra_text_loc=extra_text_loc,
            text_size=legtext_size,
            padding_loc=leg_padding,
        )

        fix_axes(
            ax1,
            ratio_axes,
            fig,
            x_ticks_ndp=x_ticks_ndp,
            yscale=yscale,
            logy=logy,
            noSci=no_sci,
            center_rlabels=center_rlabels,
        )

    if cms_label:
        add_cms_decor(ax1, cms_label, lumi=lumi, loc=logoPos)

    return fig


def plotRatio(
    ax,
    ratio_hists,
    labels,
    colors,
    linestyles=[],
    linewidth=2,
    lowerLegPos="upper right",
    lowerLegCols=2,
    legtext_size=None,
    alpha=1.0,
    yerr=False,
    fill_between=0,
    dataIdx=-1,
    baseline=True,
    add_legend=True,
    lower_leg_padding="auto",
):
    if fill_between != 0:
        for upr, downr, color in zip(
            ratio_hists[-fill_between::2],
            ratio_hists[-fill_between + 1 :: 2],
            colors[-fill_between::2],
        ):
            ax.fill_between(
                upr.axes[0].edges,
                np.append(upr.values(), upr.values()[-1]),
                np.append(downr.values(), downr.values()[-1]),
                step="post",
                color=color,
                alpha=0.5,
            )

    exclude_data = lambda x: [j for i, j in enumerate(x) if i != dataIdx]

    hep.histplot(
        exclude_data(ratio_hists)[not baseline :],
        histtype="step",
        color=exclude_data(colors)[not baseline :],
        linestyle=exclude_data(linestyles)[not baseline :],
        linewidth=linewidth,
        yerr=yerr,
        stack=False,
        ax=ax,
        alpha=alpha,
        flow="none",
    )
    if dataIdx is not None:
        hep.histplot(
            ratio_hists[dataIdx],
            histtype="errorbar",
            color=colors[dataIdx],
            xerr=False,
            yerr=True,
            stack=False,
            ax=ax,
            alpha=alpha,
            flow="none",
        )

    extra_handles = [
        Polygon(
            [[0, 0], [0, 0], [0, 0], [0, 0]],
            color=c,
            linestyle=l,
            linewidth=linewidth,
            alpha=alpha,
        )
        for c, l in zip(colors[-fill_between::2], linestyles[-fill_between::2])
    ]
    extra_labels = exclude_data(labels)[: len(extra_handles)]

    if add_legend:
        addLegend(
            ax,
            lowerLegCols,
            loc=lowerLegPos,
            text_size=legtext_size,
            extra_handles=extra_handles,
            extra_labels=extra_labels,
            custom_handlers=["stackfilled"],
            padding_loc=lower_leg_padding,
        )


def makeHistPlot2D(h2d, flow=False, **kwargs):
    if flow:
        xedges, yedges = extendEdgesByFlow(h2d)
    else:
        edges = h2d.axes.edges
        xedges = np.reshape(edges[0], len(edges[0]))
        yedges = edges[1][0]
    values = h2d.values(flow=flow)
    variances = h2d.variances(flow=flow)
    makePlot2D(values, variances, xedges, yedges, **kwargs)


def makePlot2D(
    values,
    variances=None,
    xedges=None,
    yedges=None,
    density=False,
    plot_uncertainties=False,
    xlabel="",
    ylabel="",
    zlabel="",
    colormap="RdBu",
    plot_title=None,
    ylim=None,
    xlim=None,
    zlim=None,
    zsymmetrize=None,
    logz=False,  # logy=False, logx=False, #TODO implement
    logoPos=2,
    cms_label="Work in progress",
    has_data=False,
    scaleleg=1.0,
    automatic_scale=False,
    width_scale=1.2,
):
    if xedges is None or yedges is None:
        xbins, ybins = values.shape
        if xedges is None:
            xedges = np.arange(xbins)
        if yedges is None:
            yedges = np.arange(ybins)
    # if variances is None:
    #     logger.warning("No variances given, assume")
    #     variances = values

    if density:
        xbinwidths = np.diff(xedges)
        ybinwidths = np.diff(yedges)
        binwidths = np.outer(xbinwidths, ybinwidths)
        values /= binwidths
        variances /= binwidths
    elif plot_uncertainties:
        # plot relative uncertainties instead
        values = np.sqrt(hh.relVariance(values, variances, fillOnes=True))

    if xlim is None:
        xlim = (xedges[0], xedges[-1])
    if ylim is None:
        ylim = (yedges[0], yedges[-1])

    fig, ax = figure(
        values,
        xlabel=xlabel,
        ylabel=ylabel,
        automatic_scale=automatic_scale,
        width_scale=width_scale,
        xlim=xlim,
        ylim=ylim,
    )

    if zlim is None:
        if logz:
            zmin = min(values[values > 0])  # smallest value that is not 0
        else:
            zmin = values.min()
        zmax = values.max()
        zlim = (zmin, zmax)

    # make symmetric range around value of zsymmetrize
    if zsymmetrize is not None:
        zrange = max((zmin - zsymmetrize), (zsymmetrize - zmax))
        zlim = [zsymmetrize - zrange, zsymmetrize + zrange]

    if plot_title:
        ax.text(
            1.0,
            1.003,
            plot_title,
            transform=ax.transAxes,
            fontsize=30,
            verticalalignment="bottom",
            horizontalalignment="right",
        )

    scale = max(1, np.divide(*ax.get_figure().get_size_inches()) * 0.3)
    hep.cms.label(
        ax=ax,
        lumi=None,
        fontsize=20 * scaleleg * scale,
        label=cms_label,
        data=has_data,
        loc=logoPos,
    )

    return fig


def extendEdgesByFlow(href, bin_flow_width=0.02):
    # add extra bin with bin wdith of a fraction of the total width
    all_edges = []
    for axis in href.axes:
        edges = axis.edges
        axis_range = edges[-1] - edges[0]
        if axis.traits.underflow:
            edges = np.insert(edges, 0, edges[0] - axis_range * bin_flow_width)
        if axis.traits.overflow:
            edges = np.append(edges, edges[-1] + axis_range * bin_flow_width)
        all_edges.append(edges)
    if len(all_edges) == 1:
        return all_edges[0]
    else:
        return all_edges


def fix_axes(
    ax1,
    ratio_axes=None,
    fig=None,
    x_ticks_ndp=None,
    yscale=None,
    logy=False,
    noSci=False,
    center_rlabels=False,
):
    if yscale:
        ymin, ymax = ax1.get_ylim()
        ax1.set_ylim(ymin, ymax * yscale)

    ax1.tick_params(axis="y", pad=5)  # Set distance to axis for y-axis numbers
    redo_axis_ticks(ax1, "x")

    if noSci and not logy:
        redo_axis_ticks(ax1, "y")
    elif not logy:
        ax1.ticklabel_format(style="sci", useMathText=True, axis="y", scilimits=(0, 0))

    if ratio_axes is not None:
        if not isinstance(ratio_axes, (list, tuple, np.ndarray)):
            ratio_axes = [ratio_axes]

        ax1.set_xticklabels([])
        for i, ax in enumerate(ratio_axes):
            if i == len(ratio_axes) - 1:
                redo_axis_ticks(ax, "x")
            else:
                ax.set_xticklabels([])

        # Function to get the position of the ylabel in axes coordinates
        def get_ylabel_position(ax):
            label = ax.get_yaxis().get_label()
            fig.canvas.draw()  # This is necessary to update the figure
            return ax.transAxes.inverted().transform(
                label.get_window_extent().get_points()
            )[0, 0]

        # Get the leftmost position of the y-axis labels
        y_label_pos = min(*[get_ylabel_position(ax) for ax in [ax1, *ratio_axes]])

        # Set all labels to the leftmost position
        ax1.yaxis.set_label_coords(y_label_pos * 0.7, 1.0)

        for i, ax in enumerate(ratio_axes):
            ax.tick_params(axis="y", pad=5)  # Set distance to axis for y-axis numbers
            ax.yaxis.set_label_coords(y_label_pos * 0.7, 0.5 if center_rlabels else 1.0)

        if x_ticks_ndp:
            ax.xaxis.set_major_formatter(
                StrMethodFormatter("{x:." + str(x_ticks_ndp) + "f}")
            )


def redo_axis_ticks(ax, axlabel, no_labels=False):
    autoloc = ticker.AutoLocator()
    # Need this to avoid a warning when you set the axis values manually
    fixedloc = ticker.FixedLocator(
        autoloc.tick_values(*getattr(ax, f"get_{axlabel}lim")())
    )
    getattr(ax, f"{axlabel}axis").set_major_locator(fixedloc)
    ticks = getattr(ax, f"get_{axlabel}ticks")()
    labels = [format_axis_num(x, ticks[-1]) for x in ticks] if not no_labels else []
    getattr(ax, f"set_{axlabel}ticklabels")(labels)


def format_axis_num(val, maxval):
    if type(val) == int or val.is_integer():
        # This is kinda dumb and I might change it
        return f"{val:.0f}" if maxval > 5 else f"{val:0.1f}"
    return f"{val:0.3g}" if maxval > 10 else f"{val:0.2g}"


def save_pdf_and_png(outdir, basename, fig=None):
    fname = f"{outdir}/{basename}.pdf"
    if fig:
        fig.savefig(fname, bbox_inches="tight")
        fig.savefig(fname.replace(".pdf", ".png"), bbox_inches="tight")
    else:
        plt.savefig(fname, bbox_inches="tight")
        plt.savefig(fname.replace(".pdf", ".png"), bbox_inches="tight")
    logger.info(f"Wrote file(s) {fname}(.png)")
    logger.info(f"Wrote file(s) {fname}(.png)")


def make_summary_plot(
    centerline,
    center_unc,
    center_unc_part,
    center_label,
    df,
    colors,
    xlim,
    xlabel,
    ylim=None,
    legend_loc="upper right",
    double_colors=False,
    capsize=10,
    width_scale=1.5,
    center_color="black",
    label_points=True,
    legtext_size=None,
    lumi=None,
    bbox_to_anchor=None,
    markers=None,
    top_offset=0,
    bottom_offset=0,
    padding=4,
    point_size=0.24,
    point_center_colors=None,
    cms_label="Preliminary",
    logoPos=0,
    leg_padding="auto",
):
    nentries = len(df) + (bottom_offset - top_offset)

    # This code makes me feel like an idiot by I can't think of a better way to do it
    if type(colors) == str and colors == "auto":
        cmap = mpl.cm.get_cmap("tab10")
        colors = [cmap(i) for i in range(len(df))]

    if len(colors) < len(df):
        raise ValueError(
            f"Length of values ({nentries}) must be equal or smaller than colors!"
        )
    if ylim is None:
        ylim = [0.2, nentries + 1.5]

    fig, ax1 = figure(
        None,
        xlabel=xlabel,
        ylabel="",
        grid=False,
        automatic_scale=False,
        width_scale=width_scale,
        height=padding + point_size * nentries,
        xlim=xlim,
        ylim=ylim,
    )

    ax1.plot(
        [centerline, centerline],
        ylim,
        linestyle="solid",
        marker="none",
        color=center_color,
        linewidth=2,
    )
    if center_unc_part is not None:
        ax1.fill_between(
            [centerline - center_unc_part, centerline + center_unc_part],
            *ylim,
            color="silver",
            alpha=0.6,
        )
    ax1.fill_between(
        [centerline - center_unc, centerline + center_unc],
        *ylim,
        color="silver",
        alpha=0.6,
    )

    if center_unc_part is None:
        extra_handles = [
            (
                Polygon(
                    [[0, 0], [0, 0], [0, 0], [0, 0]],
                    color="silver",
                    linestyle="solid",
                    alpha=0.6,
                ),
                Line2D(
                    [0, 0], [0, 0], color=center_color, linestyle="solid", linewidth=2
                ),
            )
        ]
    else:
        extra_handles = [
            (
                Polygon(
                    [[0, 0], [0, 0], [0, 0], [0, 0]],
                    facecolor="silver",
                    linestyle="solid",
                    edgecolor="black",
                    linewidth=2,
                    alpha=0.6,
                ),
            )
        ]

    extra_labels = [center_label]

    if markers is None:
        markers = ["o"] * len(df)

    textsize = get_textsize(ax1, legtext_size)

    for i, (x, row) in enumerate(df.iterrows()):
        # Use for spacing purposes
        vals = row.iloc[1:].values
        xpos = vals[0]
        u = vals[1:]
        pos = nentries - i - top_offset
        ax1.errorbar(
            [xpos],
            [pos],
            xerr=u[0],
            linestyle="",
            linewidth=3,
            marker=markers[i],
            color=colors[i],
            capsize=capsize,
            # label=row.loc["Name"] if label_points else None,
        )
        if len(u) > 1:
            ax1.errorbar(
                [xpos],
                [pos],
                xerr=u[1],
                linestyle="",
                linewidth=3,
                marker=markers[i],
                color=colors[i] if not point_center_colors else point_center_colors[i],
                capsize=capsize,
            )
        if label_points:
            ax1.annotate(
                row["Name"],
                (xlim[0] + 0.01 * (xlim[1] - xlim[0]), pos),
                fontsize=textsize,
                ha="left",
                va="center",
                # color="dimgrey",
                annotation_clip=False,
                style=None,
            )

    if cms_label:
        add_cms_decor(
            ax1,
            cms_label,
            loc=logoPos,
            lumi=lumi,
            no_energy=lumi is None,
            text_size=textsize,
        )

    if legend_loc is not None or bbox_to_anchor is not None:
        # Assume these are data coords, and convert to figure coords
        # Probably should be an option instead of guessing
        if bbox_to_anchor is not None and any(abs(x) > 1 for x in bbox_to_anchor):
            data_to_figure = ax1.transData + ax1.transAxes.inverted()
            bbox_to_anchor = data_to_figure.transform(bbox_to_anchor)

        addLegend(
            ax1,
            ncols=1,
            text_size=legtext_size,
            bbox_to_anchor=bbox_to_anchor,
            loc=legend_loc,
            reverse=False,
            markerfirst=True,
            labelcolor=center_color,
            extra_handles=extra_handles,
            extra_labels=extra_labels,
            extra_entries_first=False,
            custom_handlers=(
                ["verticleline"] if center_unc_part is None else ["doubleband"]
            ),
            padding_loc=leg_padding,
        )

    ax1.minorticks_off()
    ax1.set_yticklabels([])
    ax1.xaxis.set_major_locator(ticker.LinearLocator(numticks=5))

    return fig



import datetime
import glob
import os
import pathlib
import pickle
import re
import shutil
import subprocess
import sys
import tempfile
import json

import lz4.frame
import numpy as np

from wums import logging

logger = logging.child_logger(__name__)


def is_eosuser_path(path):
    if not path:
        return False
    path = os.path.realpath(path)
    return path.startswith("/eos/user") or path.startswith("/eos/home-")


def make_plot_dir(outpath, outfolder=None, eoscp=False, allowCreateLocalFolder=True):
    if eoscp and is_eosuser_path(outpath):
        # Create a unique temporary directory
        unique_temp_dir = tempfile.mkdtemp()
        outpath = os.path.join(unique_temp_dir, split_eos_path(outpath)[1])
        if not os.path.isdir(outpath):
            logger.info(f"Making temporary directory {outpath}")
            os.makedirs(outpath)

    full_outpath = outpath
    if outfolder:
        full_outpath = os.path.join(outpath, outfolder)
    if not full_outpath.endswith("/"):
        full_outpath += "/"
    if outpath and not os.path.isdir(outpath):
        # instead of raising, create folder to deal with cases where nested folders are created during code execution
        # (this would happen when outpath is already a path to a local subfolder not created in the very beginning)
        if allowCreateLocalFolder:
            logger.debug(f"Creating new directory {outpath}")
            os.makedirs(outpath)
        else:
            raise IOError(
                f"The path {outpath} doesn't not exist. You should create it (and possibly link it to your web area)"
            )

    if full_outpath and not os.path.isdir(full_outpath):
        try:
            os.makedirs(full_outpath)
            logger.info(f"Creating folder {full_outpath}")
        except FileExistsError as e:
            logger.warning(e)

    return full_outpath


def copy_to_eos(tmpFolder, outpath, outfolder=None, deleteFullTmp=False):
    eospath, outpath = split_eos_path(outpath)
    fullpath = outpath
    if outfolder:
        fullpath = os.path.join(outpath, outfolder)
    logger.info(f"Copying {tmpFolder} to {eospath}")

    for f in glob.glob(tmpFolder + "/*"):
        if not (os.path.isfile(f) or os.path.isdir(f)):
            continue
        outPathForCopy = "/".join(
            ["root://eosuser.cern.ch", eospath, f.replace(tmpFolder, f"{fullpath}/")]
        )
        if os.path.isdir(f):
            # remove last folder to do "xrdcp -fr /path/to/folder/ root://eosuser.cern.ch//eos/cms/path/to/"
            # in this way one can copy the whole subfolder through xrdcp without first creating the structure
            outPathForCopy = os.path.dirname(outPathForCopy.rstrip("/"))
        command = ["xrdcp", "-fr", f, outPathForCopy]

        logger.debug(f"Executing {' '.join(command)}")
        if subprocess.call(command):
            raise IOError(
                "Failed to copy the files to eos! Perhaps you are missing a kerberos ticket and need to run kinit <user>@CERN.CH?"
                " from lxplus you can run without eoscp and take your luck with the mount."
            )

    shutil.rmtree(tmpFolder.replace(fullpath, ""))


def split_eos_path(path):

    path = os.path.realpath(path)
    if not is_eosuser_path(path):
        raise ValueError(f"Expected a path on /eos/user, found {path}!")

    splitpath = [x for x in path.split("/") if x]
    # Can be /eos/user/<letter>/<username> or <letter-username>
    if "home-" in splitpath[1]:
        eospath = "/".join(["/eos/user", splitpath[1].split("-")[-1], splitpath[2]])
        basepath = "/".join(splitpath[3:])
    else:
        eospath = "/".join(splitpath[:4])
        basepath = "/".join(splitpath[4:])

    if path[0] == "/":
        eospath = "/" + eospath

    return eospath, basepath


def script_command_to_str(argv, parser_args):
    call_args = np.array(argv[1:], dtype=object)
    match_expr = "|".join(
        ["^-+([a-z]+[1-9]*-*)+"]
        + (
            []
            if not parser_args
            else [f"^-*{x.replace('_', '.')}" for x in vars(parser_args).keys()]
        )
    )
    if call_args.size != 0:
        flags = np.vectorize(lambda x: bool(re.match(match_expr, x)))(call_args)
        special_chars = np.vectorize(lambda x: not x.isalnum())(call_args)
        select = ~flags & special_chars
        if np.count_nonzero(select):
            call_args[select] = np.vectorize(lambda x: f"'{x}'")(call_args[select])
    return " ".join([argv[0], *call_args])


def make_meta_info_dict(
    exclude_diff="notebooks", args=None, wd=f"{pathlib.Path(__file__).parent}/../"
):
    meta_data = {
        "time": str(datetime.datetime.now()),
        "command": script_command_to_str(sys.argv, args),
        "args": {a: getattr(args, a) for a in vars(args)} if args else {},
    }
    if (
        subprocess.call(
            ["git", "branch"],
            cwd=wd,
            stderr=subprocess.STDOUT,
            stdout=open(os.devnull, "w"),
        )
        != 0
    ):
        meta_data["git_info"] = {
            "hash": "Not a git repository!",
            "diff": "Not a git repository",
        }
    else:
        meta_data["git_hash"] = subprocess.check_output(
            ["git", "log", "-1", '--format="%H"'], cwd=wd, encoding="UTF-8"
        )
        diff_comm = ["git", "diff"]
        if exclude_diff:
            diff_comm.extend(["--", f":!{exclude_diff}"])
        meta_data["git_diff"] = subprocess.check_output(
            diff_comm, encoding="UTF-8", cwd=wd
        )

    return meta_data


def write_logfile(
    outpath,
    logname,
    args={},
    meta_info={},
    wd = f"{pathlib.Path(__file__).parent}/../",
):
    logname = f"{outpath}/{logname}.log"

    with open(logname, "w") as logf:
        info = make_meta_info_dict(args=args, wd=wd)

        for k, v in {**info, **meta_info}.items():
            logf.write("\n" + "-" * 80 + "\n")
            if isinstance(v, dict):
                logf.write(k)
                logf.write(json.dumps(v, indent=5).replace("\\n", "\n"))
            else:
                logf.write(f"{k}: {v}\n")


def write_index_and_log(
    outpath,
    logname,
    template_dir=f"{pathlib.Path(__file__).parent}/Templates",
    yield_tables=None,
    analysis_meta_info=None,
    args={},
    nround=2,
    wd=f"{pathlib.Path(__file__).parent}/../",
):
    indexname = "index.php"
    if "mit.edu" in socket.gethostname() and not (
        hasattr(args, "eoscp") and args.eoscp
    ):
        indexname = "index_mit.php"

    shutil.copyfile(f"{template_dir}/{indexname}", f"{outpath}/index.php")
    logname = f"{outpath}/{logname}.log"

    with open(logname, "w") as logf:
        meta_info = (
            "-" * 80
            + "\n"
            + f"Script called at {datetime.datetime.now()}\n"
            + f"The command was: {script_command_to_str(sys.argv, args)}\n"
            + "-" * 80
            + "\n"
        )
        logf.write(meta_info)
        meta_info = make_meta_info_dict(
            "notebooks",
            args=args,
            wd=wd,
        )
        logf.write(f"git hash: {meta_info['git_hash']}\n")
        logf.write(f"git diff: {meta_info['git_diff']}\n")

        if yield_tables:
            for k, v in yield_tables.items():
                logf.write(f"Yield information for {k}\n")
                logf.write("-" * 80 + "\n")
                logf.write(str(v.round(nround)) + "\n\n")

            if (
                "Unstacked processes" in yield_tables
                and "Stacked processes" in yield_tables
            ):
                if "Data" in yield_tables["Unstacked processes"]["Process"].values:
                    unstacked = yield_tables["Unstacked processes"]
                    data_yield = unstacked[unstacked["Process"] == "Data"][
                        "Yield"
                    ].iloc[0]
                    ratio = (
                        float(
                            yield_tables["Stacked processes"]["Yield"].sum()
                            / data_yield
                        )
                        * 100
                    )
                    logf.write(f"===> Sum unstacked to data is {ratio:.2f}%")

        if analysis_meta_info:
            for k, analysis_info in analysis_meta_info.items():
                logf.write("\n" + "-" * 80 + "\n")
                logf.write(f"Meta info from input file {k}\n")
                logf.write("\n" + "-" * 80 + "\n")
                logf.write(json.dumps(analysis_info, indent=5).replace("\\n", "\n"))
        logger.info(f"Writing file {logname}")